import numpy as np
import surfa as sa
import surfa.image
# import freesurfer as fs


# surface = fs.Surface.read('/Applications/freesurfer/dev/subjects/bert/surf/lh.white')
# image = fs.Volume.read('/Applications/freesurfer/dev/subjects/bert/mri/orig.mgz')

# affine = sa.Affine(lta.matrix, source=source, target=target, transform_space='image')



# # image
# image_moved = image.transform(affine.matrix)
# image_moved.affine = 

# # surface
# affine_surf = affine.convert(space='surface')
# surface_moved = surface.copy()
# surface_moved.vertices = affine_surf.transform(surface_moved.vertices)
# surface_moved.geom = affine.target

# # ...



# aff_sv = sa.Affine(surf2vox, transform_space='surface-to-image')

# aff_sw = aff @ aff_sv

# print(aff_sw.transform_space.name)
# print(np.allclose(aff_sw.matrix, geom.surf2ras().matrix))


geom = sa.ImageGeometry((64, 64), rotation='LIA')
print(geom.rotation)
print(geom.space_transform('image', 'world')((0, 0, 0)))

geom = sa.ImageGeometry((64, 64, 64), rotation='LIA')
print(geom.rotation)
print(geom.space_transform('image', 'world')((0, 0, 0)))



# img = surfa.image.Volume(np.random.randn(64, 64, 64))
# trf = img.geom.space_transform('world', 'surface').inverse()
# print(trf.det())



from copy import deepcopy
import numpy as np

from surfa.core.array import pad_vector_length
from surfa.core.array import check_array
from surfa.transform.space import cast_space
from surfa.transform.affine import Affine
from surfa.transform.affine import cast_affine
from surfa.transform.orientation import rotation_matrix_to_orientation
from surfa.transform.orientation import orientation_to_rotation_matrix


class ImageGeometry:

    def __init__(self, shape, voxsize=None, rotation=None, center=None, shear=None, vox2world=None):
        """
        An ImageGeometry object defines the correspondence between voxel coordinates and
        world (RAS) coordinates for a given image. This correspondence can be represented
        by either a singular affine (voxel-to-world transform matrix) or a set of linear
        components (voxel scale, position, rotation, and shear). These parameters and
        corresponding coordinates transforms are appropriately recomputed upon any
        modification.

        Parameters
        ----------
        shape : array_like
            Associated 2D or 3D image shape.
        voxsize : scalar or float
            Voxel size in millimeters.
        rotation : (3, 3) float
            Voxel-to-world rotation matrix. If provide, cannot also provide `vox2world`.
        center : array_like float
            World coordinate of the image center. If provide, cannot also provide `vox2world`.
        shear : (3,) float
            Image shear. If provide, cannot also provide `vox2world`.
        vox2world : (4, 4) float or Affine
            Complete Voxel-to-world transform matrix.
        """

        # check image shape
        shape = np.array(shape)
        check_array(shape, dtype='int', ndim=1, shape=([2], [3]), name='geometry shape')

        # image geometries are always 3D by nature, but its possible they are used to represent
        # the geometry of 2D data (for example an image slice). If the original input shape is 2D,
        # then let's store this information for later, before padding the shape to 3D
        self._ndim = len(shape)
        self._valid_coord_shapes = ([2], [3]) if self._ndim == 2 else [3]

        # set image shape (this will be read-only)
        self._shape = pad_vector_length(shape, 3, 1)

        # init internal parameters
        self._voxsize = None
        self._rotation = None
        self._center = None
        self._shear = None
        self._vox2world = None

        # init internal affine transforms
        self._affines = {}

        # compute geometry
        self.update(
            voxsize=voxsize,
            rotation=rotation,
            center=center,
            shear=shear,
            vox2world=vox2world,
        )

    def update(self, voxsize=None, rotation=None, center=None, shear=None, vox2world=None):
        """
        Update specific geometry parameters. Unspecified parameters will remain the
        same unless recomputation is necessary.

        Parameters
        ----------
        voxsize : scalar or float
            Voxel size in millimeters.
        rotation : (3, 3) float
            Voxel-to-world rotation matrix. If provide, cannot also provide `vox2world`.
        center : array_like float
            World coordinate of the image center. If provide, cannot also provide `vox2world`.
        shear : (3,) float
            Image shear. If provide, cannot also provide `vox2world`.
        vox2world : (4, 4) float or Affine
            Complete Voxel-to-world transform matrix.
        """

        if voxsize is not None:
            if np.isscalar(voxsize):
                # deal with a scalar voxel size input
                voxsize = np.repeat(voxsize, 3).astype('float')
            else:
                # pad to ensure array has length of 3
                voxsize = np.array(voxsize, dtype='float')
                check_array(voxsize, ndim=1, shape=self._valid_coord_shapes, name='geometry voxsize')
                voxsize = pad_vector_length(voxsize, 3, 1, copy=False)

        if vox2world is None:

            # default voxel sizes
            if voxsize is None:
                voxsize = np.ones(3) if self.voxsize is None else self.voxsize

            # get current values if not specified
            rotation = self.rotation if rotation is None else rotation
            center = self.center if center is None else center
            shear = self.shear if shear is None else shear


            # default orientation differs for 2D
            if rotation is None:
                rotation = 'PLS' if self._ndim == 2 else 'LIA'

            if isinstance(rotation, str):
                # allow for orientation to be specified by name
                rotation = orientation_to_rotation_matrix(rotation)
            else:
                # double check rotation matrix
                rotation = np.ascontiguousarray(np.array(rotation, dtype='float'))
                check_array(rotation, ndim=2, shape=(3, 3), name='geometry rotation')

            if center is None:
                # set default world center coordinate
                center = np.zeros(3)
            else:
                # sanity checks on the center coordinate
                center = np.array(center, dtype='float')
                check_array(center, ndim=1, shape=(3,), name='geometry center')

            if shear is None:
                # set default shear
                shear = np.repeat(0.0, 3)
            else:
                # sanity checks on the shear
                shear = np.array(shear, dtype='float')
                check_array(shear, ndim=1, shape=(3,), name='geometry shear')

        else:

            # ensure that affine matrix and linear components are mutually exclusive options
            if rotation is not None:
                raise ValueError('rotation and vox2world matrix cannot both be specified when computing geometry')
            if center is not None:
                raise ValueError('center and vox2world matrix cannot both be specified when computing geometry')
            if shear is not None:
                raise ValueError('shear and vox2world matrix cannot both be specified when computing geometry')

            # compute scale, rotation, center, and shear
            vox2world = cast_affine(vox2world)
            scale, rotation, center, shear = decompose_centered_affine(self.shape, vox2world)

            # if voxsize is not provided, use the computed scale from the affine, but if voxsize has
            # been provided, make sure it's closely matching the computed scale
            if voxsize is None:
                voxsize = scale
            elif not np.allclose(scale, voxsize, atol=1e-3, rtol=0.0):
                raise ValueError(f'voxel size {voxsize} differs substantially from the computed vox2world scale {scale}')

        # now we have enough information to compute the missing affine (MOVE THIS ABOVE AS WELL)
        if vox2world is None:
            self._vox2world = compose_centered_affine(self.shape, voxsize, rotation, center, shear)
        else:
            self._vox2world = vox2world.copy()

        # set the internal parameters
        self._voxsize = voxsize
        self._rotation = rotation
        self._center = center
        self._shear = shear

        # remove all the previously cached affines (if any)
        self._affines = {}

        # it's important that we set the internal affines and parameter arrays as read-only. 
        # unfortunately, numpy does not provide a way to signal whether or not actual array
        # element data has been modified. if users in-place modify the vox2world matrix, for
        # example, we have no way of knowing to update the other affine parameters. the safest
        # bet for now is to prevent direct array manipulation, which shouldn't cause much problem.
        self._vox2world._writeable = False
        self._shape.flags.writeable = False
        self._voxsize.flags.writeable = False
        self._rotation.flags.writeable = False
        self._center.flags.writeable = False
        self._shear.flags.writeable = False

    def reshape(self, shape, copy=True):
        """
        Change the geometry image shape while preserving parameters.

        Parameters
        ----------
        shape : array_like
            Target image base shape.
        copy : bool
            Return copy if geometry already matched shape.

        Returns
        -------
        ImageGeometry
            Reshaped image geometry.
        """
        shape = pad_vector_length(shape, 3, 1, copy=False)

        # return if shapes are the same
        if np.array_equal(shape, self.shape):
            return self.copy() if copy else self

        # compute a new geometry
        geom = ImageGeometry(
            shape=shape,
            voxsize=self.voxsize,
            rotation=self.rotation,
            center=self.center,
        )
        return geom

    def copy(self):
        """
        Create a copy of the image geometry.
        """
        return deepcopy(self)

    @property
    def shape(self):
        """
        Reference image shape. This is a read-only parameter. To modify the
        shape, use the `reshape` function to compute a new geometry.
        """
        return self._shape

    @property
    def voxsize(self):
        """
        Voxel dimensions in millimeters.
        """
        return self._voxsize

    @voxsize.setter
    def voxsize(self, value):
        self.update(voxsize=value)

    @property
    def rotation(self):
        """
        Voxel-to-world rotation matrix.
        """
        return self._rotation

    @rotation.setter
    def rotation(self, value):
        self.update(rotation=value)

    @property
    def center(self):
        """
        World coordinate of image center.
        """
        return self._center

    @center.setter
    def center(self, value):
        self.update(center=value)

    @property
    def shear(self):
        """
        Image shear.
        """
        return self._shear

    @property
    def vox2world(self):
        """
        Affine transform that maps voxel (image) to world coordinates.
        """
        return self._vox2world

    @vox2world.setter
    def vox2world(self, value):
        self.update(vox2world=value, voxsize=self.voxsize)

    @property
    def world2vox(self):
        """
        Affine transform that maps world to voxel (image) coordinates.
        """
        func = lambda : self.vox2world.inv()
        return self._retrieve_or_compute_affine('wv', func)

    @world2vox.setter
    def world2vox(self, value):
        value = cast_affine(value)
        self.update(vox2world=value.inv())

    @property
    def vox2surf(self):
        """
        Affine transform that maps voxel (image) to surface coordinates.
        """
        def func():
            rot = orientation_to_rotation_matrix('LIA')
            return compose_centered_affine(self.shape, self.voxsize, rot, np.zeros(3), np.zeros(3))
        return self._retrieve_or_compute_affine('vs', func)

    @property
    def world2surf(self):
        """
        Affine transform that maps world to surface coordinates.
        """
        func = lambda : self.vox2surf @ self.world2vox
        return self._retrieve_or_compute_affine('ws', func)

    @property
    def surf2vox(self):
        """
        Affine transform that maps surface to voxel (image) coordinates.
        """
        func = lambda : self.vox2surf.inv()
        return self._retrieve_or_compute_affine('sv', func)

    @property
    def surf2world(self):
        """
        Affine transform that maps surface to world coordinates.
        """
        func = lambda : self.vox2world @ self.surf2vox
        return self._retrieve_or_compute_affine('sw', func)

    def _retrieve_or_compute_affine(self, name, func):
        """
        Internal utility to compute and cache affine matrices when called.
        """
        retrieved = self._affines.get(name)
        if retrieved is not None:
            return retrieved
        computed = func()
        computed._writeable = False
        self._affines[name] = computed
        return computed

    def affine(self, source, target):
        """
        Retrieve an affine transform based on the source and target space.

        Parameters
        ----------
        source : Space
            Source coordinate space.
        target : Space
            Target coordinate space.

        Returns
        -------
        Affine
            Retrieved affine.
        """
        a = str(cast_space(source))[0]
        b = str(cast_space(target))[0]
        # lets use the first character of the space for a quick lookup
        aff = {
            'vw': self.vox2world,
            'vs': self.vox2surf,
            'wv': self.world2vox,
            'ws': self.world2surf,
            'sv': self.surf2vox,
            'sw': self.surf2world,
        }.get(f'{a}{b}')
        if aff is None:
            raise ValueError(f'cannot find geometry affine for key {a}{b} - this is bug, not a user error')
        return aff

    @property
    def orientation(self):
        """
        Orientation string of rotation matrix.
        """
        return rotation_matrix_to_orientation(self.rotation)


def decompose_centered_affine(shape, affine):
    """
    Decompose an image-to-world affine into geometry parameters.

    Parameters
    ----------
    shape : array_like
        Shape of target image.
    affine : Affine
        Image-to-world affine to decompose.

    Returns
    -------
    tuple
        Tuple containing (voxelsize, rotation matrix, world-center, shear) parameters.
    """
    center = np.matmul(affine.matrix, np.append(np.asarray(shape) / 2, 1))[:3]
    q, r = np.linalg.qr(affine.matrix[:3, :3])
    di = np.diag_indices(3)
    voxsize = np.abs(r[di])
    p = np.eye(3)
    p[di] = r[di] / voxsize
    rotation = q @ p
    mshear = (p @ r) / np.expand_dims(voxsize, -1)
    shear = np.array([mshear[0, 1], mshear[0, 2], mshear[1, 2]])
    return (voxsize, rotation, center, shear)


def compose_centered_affine(shape, voxsize, rotation, center, shear):
    """
    Compose an image-to-world affine from geometry parameters.

    Parameters
    ----------
    shape : array_like
        Shape of target image.
    voxsize : float or tuple of floats
        Voxel size in millimeters.
    rotation : (3, 3) float or str
        Voxel-to-world rotation matrix or orientation string. If provided,
        cannot also provide `vox2world`.
    center : tuple of floats
        World coordinate of image center. If provided, cannot also provide `vox2world`.
    shear : tuple of floats
        Image shear. If provided, cannot also provide `vox2world`.

    Returns
    -------
    Affine
        Composed image-to-world  affine.
    """
    matshear = np.eye(3)
    matshear[0, 1] = shear[0]
    matshear[0, 2] = shear[1]
    matshear[1, 2] = shear[2]
    affine = np.eye(4)
    affine[:3, :3] = rotation @ (np.diag(voxsize) @ matshear)
    offset = affine @ np.append(shape / 2, 1)
    affine[:3, 3] = center - offset[:3]
    return Affine(affine)


def cast_image_geometry(obj, allow_none=True, copy=False):
    """
    Cast object to `ImageGeometry` type.

    Parameters
    ----------
    obj : any
        Object to cast.
    allow_none : bool
        Allow for `None` to be successfully passed and returned by cast.
    copy : bool
        Return copy if object is already the correct type.

    Returns
    -------
    ImageGeometry or None
        Casted image geometry.
    """
    if obj is None and allow_none:
        return obj

    if getattr(obj, '__geometry__', None) is not None:
        obj = obj.__geometry__()

    if isinstance(obj, ImageGeometry):
        return obj.copy() if copy else obj

    raise ValueError('cannot convert type %s to ImageGeometry' % type(obj).__name__)


def image_geometry_equal(a, b, tol=0.0):
    """
    Test whether two image geometries are equivalent.

    Parameters
    ----------
    a, b : ImageGeometry
        Input image geometries.
    tol : float
        Absolute error tolerance between affine matrices.

    Returns
    -------
    bool
        True if the image geometries are equal.
    """
    try:
        a = cast_image_geometry(a, allow_none=False)
        b = cast_image_geometry(b, allow_none=False)
    except ValueError:
        return False

    matches = (
        np.allclose(a.shape, b.shape, atol=tol, rtol=0.0),
        np.allclose(a.voxsize, b.voxsize, atol=tol, rtol=0.0),
        np.allclose(a.center, b.center, atol=tol, rtol=0.0),
        np.allclose(a.rotation, b.rotation, atol=tol, rtol=0.0),
        np.allclose(a.shear, b.shear, atol=tol, rtol=0.0),
        np.allclose(a.vox2world.matrix, b.vox2world.matrix, atol=tol, rtol=0.0),
    )

    if not all(matches):
        return False

    return True



import os
import numpy as np

from surfa.core.framed import FramedArray
from surfa.core.array import pad_vector_length
from surfa.core.array import check_array
from surfa.transform.geometry import ImageGeometry
from surfa.transform.geometry import cast_image_geometry
from surfa.transform.geometry import image_geometry_equal
from surfa.transform import orientation as otn
from surfa.transform.affine import cast_affine
from surfa.image.slicing import sane_slicing
from surfa.image.slicing import slicing_parameters
from surfa.image.interp import interpolate


class FramedImage(FramedArray):

    def __init__(self, basedim, data, geometry=None, **kwargs):
        """
        Abstract class defining an ND image array with data frames and associated geometry (i.e. data
        elements have a mapable relationship to a world-space coordinate system). This base class includes
        generic support for 3D and 2D FramedArray classes, which are later defined by Volume and Slice classes.

        Parameters
        ----------
        basedim : int
            Array to pad.
        data : array_like
            Internal data array.
        geometry : ImageGeometry
            Image geometry mapping the image data to world coordinates.
        metadata : dict
            Dictionary containing arbitrary array metadata.
        **kwargs
            Extra arguments provided to the FramedArray superclass.
        """
        super().__init__(basedim, data, **kwargs)
        self.geom = geometry

    def new(self, data, geometry=None):
        """
        Return a new instance of the array with updated data. Metadata and geometry is
        preserved unless specified.
        """
        geometry = geometry if geometry is not None else self.geom
        return self.__class__(data=data, geometry=geometry, metadata=self.metadata)

    @property
    def geom(self):
        """
        ImageGeometry object that positions image coordinates in 3D world space.
        """
        return self._geometry

    @geom.setter
    def geom(self, geometry):
        if geometry is None:
            geometry = ImageGeometry(shape=self.baseshape)
        else:
            geometry = cast_image_geometry(geometry).reshape(self.baseshape, copy=True)
        setattr(self, '_geometry', geometry)

    def __geometry__(self):
        """
        By setting this private function, the object can be cast to ImageGeometry.
        """
        return self.geom

    def _shape_changed(self):
        """
        Reshape geometry when the underlying shape changes.
        """
        self._geometry = self._geometry.reshape(self.baseshape)

    def smooth(self, sigma):
        """
        Smooth with Gaussian filter.

        Parameters
        ----------
        sigma : sigmascalar or sequence of scalars
            Standard deviation for Gaussian kernel.

        Returns
        -------
        FramedImage
            Smoothed image.
        """
        if np.isscalar(sigma):
            sigma = (*np.repeat(sigma, self.basdim), 0)
        else:
            sigma = np.asarray(sigma, dtype='float')
            check_array(sigma, ndim=1, shape=[[self.basedim], [self.basedim + 1]], name='sigma')
            sigma = pad_vector_length(sigma, self.basedim + 1, 0, copy=False)
        from scipy.ndimage import gaussian_filter
        return self.new(gaussian_filter(self.framed_data, sigma))

    def __getitem__(self, index_expression):
        """
        Use crop indexing similar to numpy arrays.
        """
        return self._crop(index_expression)

    def _crop(self, index_expression):
        """
        Crop the image array with a given index expression, similar to numpy
        indexing. This cropping takes into account the change in underlying image
        geometry. Certain numpy-style indexing expressions are not allowed, namely
        axis flipping (e.g. `array[::-1]`). To flip an axis, use the `reorient`
        function.

        Parameters
        ----------
        index_expression : tuple
            Numpy-style index expression.

        Returns
        -------
        FramedImage or np.ndarray
            Cropped image with updated geometry or np.ndarray if cropped result
            no longer has valid image dimensionality.
        """

        # check whether the index is valid for images, if not just return the cropped
        # data directly as a numpy ndarray
        try:
            sane_expression = sane_slicing(self.shape, index_expression)[:self.basedim]
        except IndexError:
            return self.data[index_expression]

        # extract the starting coordinate of the cropping
        start, step = slicing_parameters(sane_expression)
        start = pad_vector_length(start, 3, 0)
        step = pad_vector_length(step, 3, 1)

        if np.any(step < 1):
            raise NotImplementedError('axis flipping is not implemented via cropping, use reorient()')

        # use the original index_expression to crop the raw array
        cropped_data = self.data[index_expression]

        # compute target rotation and voxel size
        rotation = self.geom.rotation
        voxsize = self.geom.voxsize * step

        # determine if any axes are to be removed (e.g. in 3D to 2D cropping cases)
        cut_axes = [isinstance(x, int) for x in sane_expression]
        num_axis_cuts = np.count_nonzero(cut_axes)
        if (self.basedim == 2 and num_axis_cuts > 0) or (self.basedim == 3 and num_axis_cuts > 1):
            # if the array will have axes removed making it less than 2D, let's just
            # return the cropped ndarray directly, since the result is no longer an
            # image and geometry information will be invalid. In theory, a 1D array
            # could represent 3D image data along some vector, but let's not get into
            # that crazy territory
            return cropped_data
        elif num_axis_cuts == 1:
            # if we're just cutting one dimension of a 3D image, let's convert to a 2D
            # framed image, but we'll need to update the geometry information appropriately
            # to account for the change in voxel orientation
            cropped_basedim = 2
            cut_index = cut_axes.index(True)
            inter_baseshape = np.insert(cropped_data.shape, cut_index, 1)[:self.basedim]
            if cut_index < 2:
                axis_swap = [1, 2, 0] if cut_index == 0 else [0, 2, 1]
                rotation = rotation[:, axis_swap]
                voxsize = voxsize[axis_swap]
        else:
            # if a dimension is not removed, then we don't need to do much except update
            # the geometry shape and world center coordinate
            cropped_basedim = self.basedim
            inter_baseshape = np.asarray(cropped_data.shape[:self.basedim])

        # compute the new geometry
        inter_baseshape = pad_vector_length(inter_baseshape, 3, 1)
        matrix = self.geom.vox2world.matrix.copy()
        matrix[:3, 3] = self.geom.vox2world.transform(start)
        image_center = np.append(inter_baseshape * step / 2, 1)
        world_center = np.matmul(matrix, image_center)[:3]
        geometry = ImageGeometry(
            shape=cropped_data.shape[:cropped_basedim],
            center=world_center,
            rotation=rotation,
            voxsize=voxsize,
        )

        # construct cropped volume (might be Slice or Volume depending on final dimensionality)
        itype = Slice if cropped_basedim == 2 else Volume
        cropped = itype(cropped_data, geometry=geometry, metadata=self.metadata)
        return cropped

    def bbox(self, margin=None):
        """
        Compute the bounding box of the image data greater than zero. If the image
        has more than one frame, the bounding box will be computed across all frames.

        Parameters
        ----------
        margin : int of sequence of int
            Add a margin to the bounding box in units of voxels. The margin will not
            extend beyond the base image shape.

        Returns
        -------
        tuple of slice
            Bounding box as an index expression.
        """
        mask = self.max(frames=True).data > 0
        if not np.any(mask):
            return tuple([slice(0, s) for s in mask.shape])
        from scipy.ndimage import find_objects
        cropping = find_objects(mask)[0]
        if margin is not None:
            margin = np.repeat(margin, self.basedim) if np.isscalar(margin) else np.asarray(margin)
            check_array(margin, ndim=1, shape=self.basedim, name='bbox margin')
            if not np.issubdtype(margin.dtype, np.integer):
                raise ValueError('only integers can be used for valid bbox margins')
            start = [max(0, c.start - margin[i]) for i, c in enumerate(cropping)]
            stop  = [min(self.baseshape[i], c.stop + margin[i]) for i, c in enumerate(cropping)]
            step  = [c.step for c in cropping]
            cropping = tuple([slice(*s) for s in zip(start, stop, step)])
        return cropping

    def crop_to_bbox(self, margin=None):
        """
        Crop to the bounding box of image data greater than zero.

        Parameters
        ----------
        margin : int of sequence of int
            Add a margin to the bounding box in units of voxels. The margin will not
            extend beyond the base image shape.

        Returns
        -------
        FramedImage
            Cropped image with updated geometry.
        """
        return self[self.bbox(margin=margin)]

    def resize(self, voxsize, method='linear', copy=True):
        """
        Reslice image to a specified voxel size.

        Parameters
        ----------
        voxsize : scalar or float
            Voxel size in millimeters.
        method : {'linear', 'nearest'}
            Image interpolation method.
        copy : bool
            Return copy of image even if target voxel size is already satisfied.

        Returns
        -------
        FramedImage
            Resized image with updated geometry.
        """
        if self.basedim == 2:
            raise NotImplementedError('resize() is not yet implemented for 2D data, contact andrew if you need this')

        if np.isscalar(voxsize):
            # deal with a scalar voxel size input
            voxsize = np.repeat(voxsize, 3).astype('float')
        else:
            # pad to ensure array has length of 3
            voxsize = np.asarray(voxsize, dtype='float')
            check_array(voxsize, ndim=1, shape=3, name='voxsize')
            voxsize = pad_vector_length(voxsize, 3, 1, copy=False)

        # check if anything needs to be done
        if np.allclose(self.geom.voxsize, voxsize, atol=1e-5, rtol=0):
            return self.copy() if copy else self

        baseshape3D = pad_vector_length(self.baseshape, 3, 1, copy=False)
        target_shape = np.asarray(self.geom.voxsize, dtype='float') * baseshape3D / voxsize
        target_shape = tuple(np.ceil(target_shape).astype(int))

        target_geom = ImageGeometry(
            shape=target_shape,
            voxsize=voxsize,
            rotation=self.geom.rotation,
            center=self.geom.center)
        affine = self.geom.world2vox @ target_geom.vox2world
        interped = interpolate(source=self.framed_data, target_shape=target_shape,
                               method=method, affine=affine.matrix)
        return self.new(interped, target_geom)

    def resample_like(self, target, method='linear', copy=True, fill=0):
        """
        Resample to a specified target image geometry.

        Parameters
        ----------
        target : ImageGeometry
            Target image geometry to resample image data into.
        method : {'linear', 'nearest'}
            Image interpolation method.
        copy : bool
            Return copy of image even if target voxel size is already satisfied.
        fill : scalar
            Fill value for out-of-bounds voxels.

        Returns
        -------
        FramedImage
            Resampled image with updated geometry.
        """
        if self.basedim == 2:
            raise NotImplementedError('resample_like() is not yet implemented for 2D data, contact andrew if you need this')

        # cast to geometries
        source_geom = cast_image_geometry(self)
        target_geom = cast_image_geometry(target)
        if image_geometry_equal(source_geom, target_geom):
            return self.copy() if copy else self

        # compute the voxel-to-voxel affine
        affine = self.geom.world2vox @ target_geom.vox2world

        # this is an optimization to avoid interpolation if it's not needed:
        # commonly, such as when conforming images for preprocessing, images are cropped
        # to fit a given size before inputting them to some model. then, the model spits
        # out some result that must be resampled back into the original image space. however,
        # if image reshaping was the only preprocessing modification (ie. no rotation or resizing),
        # then the result does not need to be interpolated back into the target domain, it just
        # needs to be mapped back to a certain region of the grid. this section checks whether
        # that can be done by first testing if the source and target voxel sizes, rotation, and
        # shear match and if the differences in starting voxel coordinates are near-integers.
        if np.allclose(source_geom.voxsize,  target_geom.voxsize,  atol=1e-5, rtol=0.0) and \
           np.allclose(source_geom.rotation, target_geom.rotation, atol=1e-5, rtol=0.0) and \
           np.allclose(source_geom.shear,    target_geom.shear,    atol=1e-5, rtol=0.0):
            # now check if there is a integer-difference between source and target coordinates
            coord = affine.inv().transform((0, 0, 0))
            coord_rounded = coord.round()
            if np.allclose(coord, coord_rounded, atol=1e-5, rtol=0.0):
                # compute the slicing coordinates defining the matching grid regions
                target_start = coord_rounded.astype(np.int)
                source_start = np.array([0, 0, 0])
                target_stop = target_start + source_geom.shape
                source_stop = source_start + source_geom.shape

                # refine the slicing coordinate to ensure they don't exceed the target domain
                delta = np.clip(-target_start, 0, None)
                target_start += delta
                source_start += delta
                delta = np.clip(target_stop - target_geom.shape, 0, None)
                target_stop -= delta
                source_stop -= delta

                # convert to actual array slicings
                target_slicing = tuple([slice(a, b) for a, b in zip(target_start, target_stop)])
                source_slicing = tuple([slice(a, b) for a, b in zip(source_start, source_stop)])

                # place data into target shape
                target_data = np.full((*target_geom.shape, self.nframes), fill, dtype=self.dtype)
                target_data[target_slicing] = self.framed_data[source_slicing]
                return self.new(target_data, target_geom)

        # otherwise just do the standard interpolation with the computed affine
        interped = interpolate(source=self.framed_data, target_shape=target_geom.shape,
                               method=method, affine=affine.matrix, fill=fill)
        return self.new(interped, target_geom)

    def transform(self, affine=None, disp=None, method='linear', rotation='corner', resample=True, fill=0):
        """
        Apply a transform.

        Parameters
        ----------
        affine : Affine
            Affine (linear) transform to apply.
        disp : FramedImage
            Non-linear transform to apply, in the form of a displacement vector field.
        method : {'linear', 'nearest'}
            Image interpolation method if resample is enabled.
        rotation : {'corner', 'center'}
            Apply affine with rotation axis at the image corner or center.
        resample : bool
            If enabled, voxel data will be interpolated and resampled, and geometry will be set
            the target. If disabled, voxel data will not be modified, and only the geometry will
            be updated (this is not possible if a displacement field is provided).
        fill : scalar
            Fill value for out-of-bounds voxels.

        Returns
        -------
        FramedImage
            Transformed image.
        """
        if self.basedim == 2:
            raise NotImplementedError('transform() is not yet implemented for 2D data, contact andrew if you need this')

        if not resample and disp is not None:
            raise ValueError('resampling must be enabled if transforming image with displacement field')

        # if not resampling, just change the image vox2world matrix and return
        if not resample and affine is not None:
            affine = cast_affine(affine)
            if affine.source is None and affine.target is None or affine.space is None:
                raise ValueError('affine must contain source, target, and coordinate space info to transform image header')
            transformed = self.copy()
            transformed.geom.vox2world = affine.convert(space='world', source=self) @ transformed.geom.vox2world
            return transformed

        # sanity check and preprocess the affine if resampling
        matrix = None
        if affine is not None:
            affine = cast_affine(affine)
            target_geom = self.geom
            # TODO it should be assumed that the default affine space is 'voxel' when source and target are set
            if affine.space is not None:
                if affine.source is not None and affine.target is not None:
                    affine = affine.convert(space='voxel', source=self)
                    target_geom = affine.target
                elif affine.space != 'voxel':
                    raise ValueError("affine must contain source and target info if coordinate space is not 'voxel'")
            # make sure the matrix is actually inverted since we want a target to source voxel mapping for resampling
            matrix = affine.inv().matrix

        # get displacement data
        disp_data = disp.data if disp is not None else None

        # do the interpolation
        interped = interpolate(source=self.framed_data, target_shape=target_geom.shape, method=method,
                               affine=matrix, disp=disp_data,  rotation=rotation, fill=fill)
        return self.new(interped, target_geom)

    def reorient(self, orientation, copy=True):
        """
        Realigns image data and world matrix to conform to a specific slice orientation.

        Parameters
        ----------
        orientation : str
            Case-insensitive orientation string.
        copy : bool
            Return copy of image even if target orientation is already satisfied.

        Returns
        -------
        FramedImage
            Reoriented image.
        """
        if self.basedim == 2:
            raise NotImplementedError('reorient() is not yet implemented for 2D data, contact andrew if you need this')

        trg_orientation = orientation.upper()
        src_orientation = otn.rotation_matrix_to_orientation(self.geom.vox2world.matrix)
        if trg_orientation == src_orientation.upper():
            return self.copy() if copy else self

        # extract world axes
        get_world_axes = lambda aff: np.argmax(np.absolute(np.linalg.inv(aff)), axis=0)
        trg_matrix = otn.orientation_to_rotation_matrix(trg_orientation)
        src_matrix = otn.orientation_to_rotation_matrix(src_orientation)
        world_axes_trg = get_world_axes(trg_matrix[:self.basedim, :self.basedim])
        world_axes_src = get_world_axes(src_matrix[:self.basedim, :self.basedim])

        voxsize = np.asarray(self.geom.voxsize)
        voxsize = voxsize[world_axes_src][world_axes_trg]

        # initialize new
        data = self.data.copy()
        affine = self.geom.vox2world.matrix.copy()

        # align axes
        affine[:, world_axes_trg] = affine[:, world_axes_src]
        for i in range(self.basedim):
            if world_axes_src[i] != world_axes_trg[i]:
                data = np.swapaxes(data, world_axes_src[i], world_axes_trg[i])
                swapped_axis_idx = np.where(world_axes_src == world_axes_trg[i])
                world_axes_src[swapped_axis_idx], world_axes_src[i] = world_axes_src[i], world_axes_src[swapped_axis_idx]

        # align directions
        dot_products = np.sum(affine[:3, :3] * trg_matrix[:3, :3], axis=0)
        for i in range(self.basedim):
            if dot_products[i] < 0:
                data = np.flip(data, axis=i)
                affine[:, i] = - affine[:, i]
                affine[:3, 3] = affine[:3, 3] - affine[:3, i] * (data.shape[i] - 1)

        # update geometry
        target_geom = ImageGeometry(
            shape=data.shape[:3],
            vox2world=affine,
            voxsize=voxsize)
        return self.new(data, target_geom)

    def reshape(self, shape, copy=True):
        """
        Returns a volume fit to a given shape. Image will be centered in the conformed volume.

        Parameters
        ----------
        shape : tuple of int
            Target shape.
        copy : bool
            Return copy of image even if target shape is already satisfied.

        Returns
        -------
        FramedImage
            Reshaped image.
        """
        if self.basedim == 2:
            raise NotImplementedError('reshape() is not yet implemented for 2D data, contact andrew if you need this')
        if self.nframes > 1:
            # TODO
            raise NotImplementedError

        if np.array_equal(self.baseshape, shape):
            return self.copy() if copy else self

        delta = (np.array(shape) - np.array(self.baseshape)) / 2
        low = np.floor(delta).astype(int)
        high = np.ceil(delta).astype(int)

        c_low = np.clip(low, 0, None)
        c_high = np.clip(high, 0, None)
        conformed_data = np.pad(self.data.squeeze(), list(zip(c_low, c_high)), mode='constant')

        # note: low and high are intentionally swapped here
        c_low = np.clip(-high, 0, None)
        c_high = conformed_data.shape[:3] - np.clip(-low, 0, None)
        cropping = tuple([slice(a, b) for a, b in zip(c_low, c_high)])
        conformed_data = conformed_data[cropping]

        # compute new affine if one exists
        matrix = np.eye(4)
        matrix[:3, :3] = self.geom.vox2world.matrix[:3, :3]
        p0crs = np.clip(-high, 0, None) - np.clip(low, 0, None)
        p0 = self.geom.vox2world(p0crs)
        matrix[:3, 3] = p0
        pcrs = np.append(np.array(conformed_data.shape[:3]) / 2, 1)
        cras = np.matmul(matrix, pcrs)[:3]
        matrix[:3, 3] = 0
        matrix[:3, 3] = cras - np.matmul(matrix, pcrs)[:3]

        # update geometry
        target_geom = ImageGeometry(
            shape=conformed_data.shape[:3],
            vox2world=matrix,
            voxsize=self.geom.voxsize)
        return self.new(conformed_data, target_geom)

    def conform(self, shape=None, voxsize=1.0, orientation='LIA', method='linear', dtype=None, copy=True):
        """
        Conforms image to a specific shape, type, resolution, and orientation.

        Parameters
        ----------
        shape : tuple of int
            Target shape.
        voxsize : scalar or float
            Voxel size in millimeters.
        orientation : str
            Case-insensitive orientation string.
        method : {'linear', 'nearest'}
            Image interpolation method if resample is enabled.
        dtype : np.dtype
            Target datatype.
        copy : bool
            Return copy of image even if requirements are already satisfied.

        Returns
        -------
        FramedImage
            Conformed image.
        """
        conformed = self.reorient(orientation, copy=False)
        conformed = conformed.resize(voxsize, method=method, copy=False)
        if shape is not None:
            conformed = conformed.reshape(shape, copy=False)
        if dtype is not None:
            conformed = conformed.astype(dtype, copy=False)
        return self.copy() if (copy and conformed is self) else conformed


class Slice(FramedImage):

    def __init__(self, data, **kwargs):
        super().__init__(basedim=2, data=data, **kwargs)


class Volume(FramedImage):

    def __init__(self, data, **kwargs):
        super().__init__(basedim=3, data=data, **kwargs)


def cast_image(obj, allow_none=True, copy=False):
    """
    Cast object to `FramedImage` type.

    Parameters
    ----------
    obj : any
        Object to cast.
    allow_none : bool
        Allow for `None` to be successfully passed and returned by cast.
    copy : bool
        Return copy if object is already the correct type.

    Returns
    -------
    FramedImage or None
        Casted image.
    """
    if obj is None and allow_none:
        return obj

    if isinstance(obj, FramedImage):
        return obj.copy() if copy else obj

    if getattr(obj, '__array__', None) is not None:
        return Volume(np.array(obj))

    # as a final test, check if the input is possibly a nibabel image
    # we don't want nibabel to be required though, so ignore import errors
    try:
        import nibabel as nib
        if isinstance(obj, nib.spatialimages.SpatialImage):
            data = obj.get_data().copy()
            geometry = ImageGeometry(data.shape[:3], vox2world=obj.affine)
            return Volume(data, geometry=geometry)
    except ImportError:
        pass

    raise ValueError('cannot convert type %s to image' % type(obj).__name__)



import numpy as np


def rotation_matrix_to_orientation(matrix):
    """
    Examines a (3, 3) direction cosine matrix and determines the corresponding orientation,
    indicating the primary direction of each axis in the 3D matrix. The characters can
    be L, R, A, P, I, S. Case is not important, but uppercase is generally used.

    Parameters
    ----------
    matrix : (3, 3) float
        Direction cosine matrix.

    Returns
    -------
        str: Computed orientation string.
    """
    matrix = matrix[:3, :3]
    orientation = ''
    for i in range(3):
        sag, cor, ax = matrix[:, i]
        if np.abs(sag) > np.abs(cor) and np.abs(sag) > np.abs(ax):
            orientation += 'R' if sag > 0 else 'L'
        elif np.abs(cor) > np.abs(ax):
            orientation += 'A' if cor > 0 else 'P'
        else:
            orientation += 'S' if ax > 0 else 'I'
    return orientation


def orientation_to_rotation_matrix(orientation):
    """
    Computes the direction cosine matrix corresponding to an orientation.

    Parameters
    ----------
    orientation : str
        Case-insensitive orientation string.

    Returns
    -------
    np.ndarray
        4x4 direction cosine matrix.
    """
    orientation = orientation.upper()
    check_orientation(orientation)

    matrix = np.zeros((3, 3))
    for i, c in enumerate(orientation):
        matrix[:3, i] -= [c == x for x in 'LPI']
        matrix[:3, i] += [c == x for x in 'RAS']
    return matrix


def check_orientation(orientation):
    """
    Checks an orientation string to ensure it is valid, meaning that all axes are represented
    exactly once and no invalid characters are present.

    Parameters
    ----------
    orientation : str
        Case-insensitive orientation string.
    """
    orientation = orientation.upper()

    if len(orientation) != 3:
        raise ValueError('bad orientation: expected 3 characters, got %d' % len(orientation))

    axes = ['LR', 'PA', 'IS']
    rs = np.zeros(3, dtype='int')
    for c in orientation:
        r = [c in axis for axis in axes]
        if not any(r):
            raise ValueError("bad orientation: unknown character '{c}'")
        rs += r

    for i in range(3):
        if rs[i] > 1:
            raise ValueError('bad orientation: %s axis represented multiple times' % axes[i])
        if rs[i] == 0:
            raise ValueError('bad orientation: %s axis not represented' % axes[i])


def slice_direction(orientation):
    """
    Determines primary slice direction string from orientation.

    Parameters
    ----------
    orientation : str
        Case-insensitive orientation string.

    Returns
    -------
    str
        Primary slice direction.
    """
    check_orientation(orientation)
    if orientation[2] in 'LR':
        return 'sagittal'
    if orientation[2] in 'PA':
        return 'coronal'
    if orientation[2] in 'IS':
        return 'axial'



import sys
import numpy as np

cimport cython
cimport numpy as np

from libc.math cimport floor
from libc.math cimport round


def interpolate(source, target_shape, method, affine=None, disp=None, rotation='corner', fill=0):
    """
    Interpolate a 3D image given a voxel-to-voxel affine transform and/or a
    dense displacement field.

    Parameters
    ----------
    source : array_like
        4-dimensional source numpy array, with the last dimension representing data frames.
    target_shape : tuple of ints
        Target base shape of interpolated output. Must be a 3D shape.
    method : str
        Interpolation method. Must 'linear' or 'nearest'.
    affine : array_like
        Square affine transform that maps target voxels coordinates to source voxel coordinates.
    disp : array_like
        Dense vector displacement field. Base shape must match target shape.
    rotation : str
        Origin of image rotation. Must be 'corner' or 'center'.
    fill : scalar
        Fill value for out-of-bounds voxels.

    Returns
    -------
    np.ndarray
        Interpolated image array.
    """
    if affine is None and disp is None:
        raise ValueError('interpolation requires an affine transform and/or displacement field')

    if method not in ('linear', 'nearest'):
        raise ValueError(f'interp method must be linear or nearest, but got {method}')

    if not isinstance(source, np.ndarray):
        raise ValueError(f'source data must a numpy array, but got input of type {source.__class__.__name__}')

    if source.ndim != 4:
        raise ValueError(f'source data must be 4D, but got input of shape {target_shape}')

    target_shape = tuple(target_shape)
    if len(target_shape) != 3:
        raise ValueError(f'interpolated target shape must be 3D, but got {target_shape}')

    if rotation not in ('center', 'corner'):
        raise ValueError("rotation must be 'center' or 'corner'")

    # check affine
    use_affine = affine is not None
    if use_affine:
        if not isinstance(affine, np.ndarray):
            raise ValueError(f'affine must a numpy array, but got input of type {source.__class__.__name__}')
        if not np.array_equal(affine.shape, (4, 4)):
            raise ValueError(f'affine must be 4x4, but got input of shape {affine.shape}')
        # ensure the rotation is around the image corner when applying the affine
        if rotation == 'center':
            center = np.eye(4, dtype=np.float32)
            center[:3, -1] = -0.5 * (np.asarray(source.shape[:3]) - 1)
            affine = np.linalg.inv(center) @ (affine @ center)
        # only supports float32 affines for now
        affine = affine.astype(np.float32, copy=False)

    # check displacement
    use_disp = disp is not None
    if use_disp:
        if not isinstance(disp, np.ndarray):
            raise ValueError(f'source data must a numpy array, but got input of type {source.__class__.__name__}')
        if not np.array_equal(disp.shape, target_shape):
            raise ValueError(f'displacement field shape {disp.shape} must match target shape {target_shape}')

    if not source.flags.c_contiguous and not source.flags.f_contiguous:
        # TODO figure out what would cause this
        source = np.asarray(source, order='F')

    # find corresponding function
    order = 'contiguous' if source.flags.c_contiguous else 'fortran'
    interp_func = globals().get(f'interp_3d_{order}_{method}')

    # speeds up if conditionals are computed outside of function (TODO is this even true?)
    shape = np.asarray(target_shape).astype('int64')

    # ensure correct byteorder
    # TODO maybe this should be done at read-time?
    swap_byteorder = sys.byteorder == 'little' and '>' or '<'
    source = source.byteswap().newbyteorder() if source.dtype.byteorder == swap_byteorder else source

    resampled = interp_func(source, shape, affine, disp, fill, use_affine, use_disp)
    return resampled


# data types to compile for
ctypedef fused datatype:
    cython.char
    cython.uchar
    cython.short
    cython.ushort
    cython.int
    cython.uint
    cython.long
    cython.ulong
    cython.float
    cython.double


@cython.boundscheck(False)
@cython.wraparound(False)
def interp_3d_fortran_nearest(datatype[::1, :, :, :] source,
                              np.ndarray[np.int_t, ndim=1] target_shape,
                              float[:, ::1] mat,
                              float[::1, :, :, :] disp,
                              datatype fill_value,
                              bint use_affine,
                              bint use_disp):

    # dimensions of the source image
    cdef Py_ssize_t sx_max = source.shape[0]
    cdef Py_ssize_t sy_max = source.shape[1]
    cdef Py_ssize_t sz_max = source.shape[2]
    cdef Py_ssize_t frames = source.shape[3]

    # target image
    cdef Py_ssize_t x_max = target_shape[0]
    cdef Py_ssize_t y_max = target_shape[1]
    cdef Py_ssize_t z_max = target_shape[2]

    # fill value
    cdef datatype fill = fill_value

    # intermediate variables
    cdef Py_ssize_t x, y, z, f
    cdef float v
    cdef float sx, sy, sz
    cdef float ix, iy, iz
    cdef Py_ssize_t sx_idx, sy_idx, sz_idx

    # allocate the target image
    if   datatype is cython.char:   np_type = np.int8
    elif datatype is cython.uchar:  np_type = np.uint8
    elif datatype is cython.short:  np_type = np.int16
    elif datatype is cython.ushort: np_type = np.uint16
    elif datatype is cython.int:    np_type = np.int32
    elif datatype is cython.uint:   np_type = np.uint32
    elif datatype is cython.long:   np_type = np.int64
    elif datatype is cython.ulong:  np_type = np.uint64
    elif datatype is cython.float:  np_type = np.float32
    elif datatype is cython.double: np_type = np.float64

    target = np.zeros([x_max, y_max, z_max, frames], dtype=np_type, order='F')
    cdef datatype[::1, :, :, :] target_view = target

    # extract affine matrix values
    cdef float mat00, mat01, mat02, mat03
    cdef float mat10, mat11, mat12, mat13
    cdef float mat20, mat21, mat22, mat23
    if use_affine:
        mat00 = mat[0, 0]
        mat01 = mat[0, 1]
        mat02 = mat[0, 2]
        mat03 = mat[0, 3]
        mat10 = mat[1, 0]
        mat11 = mat[1, 1]
        mat12 = mat[1, 2]
        mat13 = mat[1, 3]
        mat20 = mat[2, 0]
        mat21 = mat[2, 1]
        mat22 = mat[2, 2]
        mat23 = mat[2, 3]

    # loop over each voxel in the target image
    for z in range(z_max):
        for y in range(y_max):
            for x in range(x_max):

                # transform the current target coordinate to get
                # the point in source space
                if use_disp:
                    ix = x + disp[x, y, z, 0]
                    iy = y + disp[x, y, z, 1]
                    iz = z + disp[x, y, z, 2]
                else:
                    ix = x
                    iy = y
                    iz = z

                if use_affine:
                    sx = (mat00 * ix) + (mat01 * iy) + (mat02 * iz) + mat03
                    sy = (mat10 * ix) + (mat11 * iy) + (mat12 * iz) + mat13
                    sz = (mat20 * ix) + (mat21 * iy) + (mat22 * iz) + mat23
                else:
                    sx = ix
                    sy = iy
                    sz = iz

                # check coordinate limits
                if sx < 0 or sx >= sx_max or \
                   sy < 0 or sy >= sy_max or \
                   sz < 0 or sz >= sz_max:
                    for f in range(frames):
                        target_view[x, y, z, f] = fill
                    continue

                # round to nearest voxel
                sx_idx = int(round(sx))
                sy_idx = int(round(sy))
                sz_idx = int(round(sz))
                if sx_idx == sx_max: sx_idx -= 1
                if sy_idx == sy_max: sy_idx -= 1
                if sz_idx == sz_max: sz_idx -= 1

                # sample each frame
                for f in range(frames):
                    target_view[x, y, z, f] = source[sx_idx, sy_idx, sz_idx, f]

    return target


@cython.boundscheck(False)
@cython.wraparound(False)
def interp_3d_fortran_linear(datatype[::1, :, :, :] source,
                             np.ndarray[np.int_t, ndim=1] target_shape,
                             float[:, ::1] mat,
                             float[::1, :, :, :] disp,
                             datatype fill_value,
                             bint use_affine,
                             bint use_disp):

    # dimensions of the source image
    cdef Py_ssize_t sx_max_idx = source.shape[0] - 1
    cdef Py_ssize_t sy_max_idx = source.shape[1] - 1
    cdef Py_ssize_t sz_max_idx = source.shape[2] - 1
    cdef Py_ssize_t frames = source.shape[3]

    # target image
    cdef Py_ssize_t x_max = target_shape[0]
    cdef Py_ssize_t y_max = target_shape[1]
    cdef Py_ssize_t z_max = target_shape[2]

    # fill value
    cdef float fill = fill_value

    # intermediate variables
    cdef Py_ssize_t x, y, z, f
    cdef float v
    cdef float sx, sy, sz
    cdef float ix, iy, iz
    cdef Py_ssize_t sx_low, sy_low, sz_low
    cdef Py_ssize_t sx_high, sy_high, sz_high
    cdef float dsx, dsy, dsz
    cdef float w0, w1, w2, w3, w4, w5, w6, w7

    # allocate the target image
    target = np.zeros([x_max, y_max, z_max, frames], dtype=np.float32, order='F')
    cdef float[::1, :, :, :] target_view = target

    # extract affine matrix values
    cdef float mat00, mat01, mat02, mat03
    cdef float mat10, mat11, mat12, mat13
    cdef float mat20, mat21, mat22, mat23
    if use_affine:
        mat00 = mat[0, 0]
        mat01 = mat[0, 1]
        mat02 = mat[0, 2]
        mat03 = mat[0, 3]
        mat10 = mat[1, 0]
        mat11 = mat[1, 1]
        mat12 = mat[1, 2]
        mat13 = mat[1, 3]
        mat20 = mat[2, 0]
        mat21 = mat[2, 1]
        mat22 = mat[2, 2]
        mat23 = mat[2, 3]

    # loop over each voxel in the target image
    for z in range(z_max):
        for y in range(y_max):
            for x in range(x_max):

                # transform the current target coordinate to get
                # the point in source space
                if use_disp:
                    ix = x + disp[x, y, z, 0]
                    iy = y + disp[x, y, z, 1]
                    iz = z + disp[x, y, z, 2]
                else:
                    ix = x
                    iy = y
                    iz = z

                if use_affine:
                    sx = (mat00 * ix) + (mat01 * iy) + (mat02 * iz) + mat03
                    sy = (mat10 * ix) + (mat11 * iy) + (mat12 * iz) + mat13
                    sz = (mat20 * ix) + (mat21 * iy) + (mat22 * iz) + mat23
                else:
                    sx = ix
                    sy = iy
                    sz = iz

                # get low and high coords
                sx_low = int(floor(sx))
                sy_low = int(floor(sy))
                sz_low = int(floor(sz))

                # check coordinate limits
                if sx_low < 0 or sx_low > sx_max_idx or \
                   sy_low < 0 or sy_low > sy_max_idx or \
                   sz_low < 0 or sz_low > sz_max_idx:
                    for f in range(frames):
                        target_view[x, y, z, f] = fill
                    continue

                # make sure high value does not exceed limit
                sx_high = sx_low
                sy_high = sy_low
                sz_high = sz_low
                if sx_low != sx_max_idx:
                    sx_high += 1
                if sy_low != sy_max_idx:
                    sy_high += 1
                if sz_low != sz_max_idx:
                    sz_high += 1

                # get coordinate diff
                sx -= sx_low
                sy -= sy_low
                sz -= sz_low
                dsx = 1.0 - sx
                dsy = 1.0 - sy
                dsz = 1.0 - sz

                # compute weights
                w0 = dsx * dsy * dsz;
                w1 = sx  * dsy * dsz;
                w2 = dsx * sy  * dsz;
                w3 = dsx * dsy * sz;
                w4 = sx  * dsy * sz;
                w5 = dsx * sy  * sz;
                w6 = sx  * sy  * dsz;
                w7 = sx  * sy  * sz;

                # interpolate for each frame
                for f in range(frames):
                    v = w0 * source[sx_low , sy_low , sz_low , f] + \
                        w1 * source[sx_high, sy_low , sz_low , f] + \
                        w2 * source[sx_low , sy_high, sz_low , f] + \
                        w3 * source[sx_low , sy_low , sz_high, f] + \
                        w4 * source[sx_high, sy_low , sz_high, f] + \
                        w5 * source[sx_low , sy_high, sz_high, f] + \
                        w6 * source[sx_high, sy_high, sz_low , f] + \
                        w7 * source[sx_high, sy_high, sz_high, f]
                    target_view[x, y, z, f] = v

    return target


@cython.boundscheck(False)
@cython.wraparound(False)
def interp_3d_contiguous_nearest(datatype[:, :, :, ::1] source,
                                 np.ndarray[np.int_t, ndim=1] target_shape,
                                 float[:, ::1] mat,
                                 float[:, :, :, ::1] disp,
                                 datatype fill_value,
                                 bint use_affine,
                                 bint use_disp):

    # dimensions of the source image
    cdef Py_ssize_t sx_max = source.shape[0]
    cdef Py_ssize_t sy_max = source.shape[1]
    cdef Py_ssize_t sz_max = source.shape[2]
    cdef Py_ssize_t frames = source.shape[3]

    # target image
    cdef Py_ssize_t x_max = target_shape[0]
    cdef Py_ssize_t y_max = target_shape[1]
    cdef Py_ssize_t z_max = target_shape[2]

    # fill value
    cdef datatype fill = fill_value

    # intermediate variables
    cdef Py_ssize_t x, y, z, f
    cdef float v
    cdef float sx, sy, sz
    cdef float ix, iy, iz
    cdef Py_ssize_t sx_idx, sy_idx, sz_idx

    # allocate the target image
    if   datatype is cython.char:   np_type = np.int8
    elif datatype is cython.uchar:  np_type = np.uint8
    elif datatype is cython.short:  np_type = np.int16
    elif datatype is cython.ushort: np_type = np.uint16
    elif datatype is cython.int:    np_type = np.int32
    elif datatype is cython.uint:   np_type = np.uint32
    elif datatype is cython.long:   np_type = np.int64
    elif datatype is cython.ulong:  np_type = np.uint64
    elif datatype is cython.float:  np_type = np.float32
    elif datatype is cython.double: np_type = np.float64

    target = np.zeros([x_max, y_max, z_max, frames], dtype=np_type, order='F')
    cdef datatype[::1, :, :, :] target_view = target

    # extract affine matrix values
    cdef float mat00, mat01, mat02, mat03
    cdef float mat10, mat11, mat12, mat13
    cdef float mat20, mat21, mat22, mat23
    if use_affine:
        mat00 = mat[0, 0]
        mat01 = mat[0, 1]
        mat02 = mat[0, 2]
        mat03 = mat[0, 3]
        mat10 = mat[1, 0]
        mat11 = mat[1, 1]
        mat12 = mat[1, 2]
        mat13 = mat[1, 3]
        mat20 = mat[2, 0]
        mat21 = mat[2, 1]
        mat22 = mat[2, 2]
        mat23 = mat[2, 3]

    # loop over each voxel in the target image
    for x in range(x_max):
        for y in range(y_max):
            for z in range(z_max):

                # transform the current target coordinate to get
                # the point in source space
                if use_disp:
                    ix = x + disp[x, y, z, 0]
                    iy = y + disp[x, y, z, 1]
                    iz = z + disp[x, y, z, 2]
                else:
                    ix = x
                    iy = y
                    iz = z

                if use_affine:
                    sx = (mat00 * ix) + (mat01 * iy) + (mat02 * iz) + mat03
                    sy = (mat10 * ix) + (mat11 * iy) + (mat12 * iz) + mat13
                    sz = (mat20 * ix) + (mat21 * iy) + (mat22 * iz) + mat23
                else:
                    sx = ix
                    sy = iy
                    sz = iz

                # check coordinate limits
                if sx < 0 or sx >= sx_max or \
                   sy < 0 or sy >= sy_max or \
                   sz < 0 or sz >= sz_max:
                    for f in range(frames):
                        target_view[x, y, z, f] = fill
                    continue

                # round to nearest voxel
                sx_idx = int(round(sx))
                sy_idx = int(round(sy))
                sz_idx = int(round(sz))
                if sx_idx == sx_max: sx_idx -= 1
                if sy_idx == sy_max: sy_idx -= 1
                if sz_idx == sz_max: sz_idx -= 1

                # sample each frame
                for f in range(frames):
                    target_view[x, y, z, f] = source[sx_idx, sy_idx, sz_idx, f]

    return target


@cython.boundscheck(False)
@cython.wraparound(False)
def interp_3d_contiguous_linear(datatype[:, :, :, ::1] source,
                                np.ndarray[np.int_t, ndim=1] target_shape,
                                float[:, ::1] mat,
                                float[:, :, :, ::1] disp,
                                datatype fill_value,
                                bint use_affine,
                                bint use_disp):

    # dimensions of the source image
    cdef Py_ssize_t sx_max_idx = source.shape[0] - 1
    cdef Py_ssize_t sy_max_idx = source.shape[1] - 1
    cdef Py_ssize_t sz_max_idx = source.shape[2] - 1
    cdef Py_ssize_t frames = source.shape[3]

    # target image
    cdef Py_ssize_t x_max = target_shape[0]
    cdef Py_ssize_t y_max = target_shape[1]
    cdef Py_ssize_t z_max = target_shape[2]

    # fill value
    cdef float fill = fill_value

    # intermediate variables
    cdef Py_ssize_t x, y, z, f
    cdef float v
    cdef float sx, sy, sz
    cdef float ix, iy, iz
    cdef Py_ssize_t sx_low, sy_low, sz_low
    cdef Py_ssize_t sx_high, sy_high, sz_high
    cdef float dsx, dsy, dsz
    cdef float w0, w1, w2, w3, w4, w5, w6, w7

    # allocate the target image
    target = np.zeros([x_max, y_max, z_max, frames], dtype=np.float32, order='F')
    cdef float[::1, :, :, :] target_view = target

    # extract affine matrix values
    cdef float mat00, mat01, mat02, mat03
    cdef float mat10, mat11, mat12, mat13
    cdef float mat20, mat21, mat22, mat23
    if use_affine:
        mat00 = mat[0, 0]
        mat01 = mat[0, 1]
        mat02 = mat[0, 2]
        mat03 = mat[0, 3]
        mat10 = mat[1, 0]
        mat11 = mat[1, 1]
        mat12 = mat[1, 2]
        mat13 = mat[1, 3]
        mat20 = mat[2, 0]
        mat21 = mat[2, 1]
        mat22 = mat[2, 2]
        mat23 = mat[2, 3]

    # loop over each voxel in the target image
    for x in range(x_max):
        for y in range(y_max):
            for z in range(z_max):

                # transform the current target coordinate to get
                # the point in source space
                if use_disp:
                    ix = x + disp[x, y, z, 0]
                    iy = y + disp[x, y, z, 1]
                    iz = z + disp[x, y, z, 2]
                else:
                    ix = x
                    iy = y
                    iz = z

                if use_affine:
                    sx = (mat00 * ix) + (mat01 * iy) + (mat02 * iz) + mat03
                    sy = (mat10 * ix) + (mat11 * iy) + (mat12 * iz) + mat13
                    sz = (mat20 * ix) + (mat21 * iy) + (mat22 * iz) + mat23
                else:
                    sx = ix
                    sy = iy
                    sz = iz

                # get low and high coords
                sx_low = int(floor(sx))
                sy_low = int(floor(sy))
                sz_low = int(floor(sz))

                # check coordinate limits
                if sx_low < 0 or sx_low > sx_max_idx or \
                   sy_low < 0 or sy_low > sy_max_idx or \
                   sz_low < 0 or sz_low > sz_max_idx:
                    for f in range(frames):
                        target_view[x, y, z, f] = fill
                    continue

                # make sure high value does not exceed limit
                sx_high = sx_low
                sy_high = sy_low
                sz_high = sz_low
                if sx_low != sx_max_idx:
                    sx_high += 1
                if sy_low != sy_max_idx:
                    sy_high += 1
                if sz_low != sz_max_idx:
                    sz_high += 1

                # get coordinate diff
                sx -= sx_low
                sy -= sy_low
                sz -= sz_low
                dsx = 1.0 - sx
                dsy = 1.0 - sy
                dsz = 1.0 - sz

                # compute weights
                w0 = dsx * dsy * dsz;
                w1 = sx  * dsy * dsz;
                w2 = dsx * sy  * dsz;
                w3 = dsx * dsy * sz;
                w4 = sx  * dsy * sz;
                w5 = dsx * sy  * sz;
                w6 = sx  * sy  * dsz;
                w7 = sx  * sy  * sz;

                # interpolate for each frame
                for f in range(frames):
                    v = w0 * source[sx_low , sy_low , sz_low , f] + \
                        w1 * source[sx_high, sy_low , sz_low , f] + \
                        w2 * source[sx_low , sy_high, sz_low , f] + \
                        w3 * source[sx_low , sy_low , sz_high, f] + \
                        w4 * source[sx_high, sy_low , sz_high, f] + \
                        w5 * source[sx_low , sy_high, sz_high, f] + \
                        w6 * source[sx_high, sy_high, sz_low , f] + \
                        w7 * source[sx_high, sy_high, sz_high, f]
                    target_view[x, y, z, f] = v

    return target



import os
import numpy as np


def check_file_readability(filename):
    """
    Raise an exception if a file cannot be read.

    Parameters
    ----------
    filename : str
        Path to file.
    """
    if os.path.isdir(filename):
        raise ValueError(f'{filename} is a directory, not a file')
    if not os.path.isfile(filename):
        raise FileNotFoundError(f'{filename} is not a file')
    if not os.access(filename, os.R_OK):
        raise PermissionError(f'{filename} is not a readable file')


def read_int(file, size=4, signed=True, byteorder='big'):
    """
    Read integer from a file buffer.

    Parameters
    ----------
    file : BufferedReader
        Opened file buffer.
    size : int
        Byte size.
    signed : bool
        Whether integer is signed.
    byteorder : str
        Memory byte order.

    Returns
    -------
    integer : int
    """
    return int.from_bytes(file.read(size), byteorder=byteorder, signed=signed)


def write_int(file, value, size=4, signed=True, byteorder='big'):
    """
    Write integer to a file buffer.

    Parameters
    ----------
    file : BufferedWriter
        Opened file buffer.
    size : int
        Byte size.
    signed : bool
        Whether integer is signed.
    byteorder : str
        Memory byte order.
    """
    file.write(value.to_bytes(size, byteorder=byteorder, signed=signed))


def read_bytes(file, dtype, count=1):
    """
    Read from a binary file buffer.

    Parameters
    ----------
    file : BufferedReader
        Opened file buffer.
    dtype : np.dtype
        Read into numpy datatype.
    count : int
        Number of elements to read.

    Returns
    -------
    np.ndarray:
        The read dtype array.
    """
    dtype = np.dtype(dtype)
    value = np.fromstring(file.read(dtype.itemsize * count), dtype=dtype)
    if count == 1:
        return value[0]
    return value


def write_bytes(file, value, dtype):
    """
    Write a binary file buffer.

    Parameters
    ----------
    file : BufferedWriter
        Opened file buffer.
    value : array_like
        Data to write.
    dtype : np.dtype
        Datatype to save as.
    """
    file.write(np.asarray(value).astype(dtype, copy=False).tobytes())



import numpy as np


def conform_ndim(arr, ndim):
    """
    Conform array to a particular dimensionality by appending empty axes.

    Parameters
    ----------
    arr : array_like
        Array to conform.
    ndim : int
        Target dimensionality of array.

    Returns
    -------
    np.ndarray
        Conformed array.
    """
    arr = np.asarray(arr)
    if arr.ndim > ndim:
        raise ValueError(f'cannot conform array of shape {arr.shape} to {ndmi}D')
    for _ in range(ndim - arr.ndim):
        arr = np.expand_dims(arr, axis=-1)
    return arr


def pad_vector_length(arr, length, fill, copy=True):
    """
    Pad a 1D vector to a particular length.

    Parameters
    ----------
    arr : array_like
        Array to pad.
    length : int
        Target length of array.
    fill : scalar
        All added elements will be assigned this value.
    copy : bool
        Return copy if input vector has correct length.

    Returns
    -------
    np.ndarray
        Padded array.
    """
    arr = np.asarray(arr)
    if arr.ndim != 1:
        raise ValueError(f'array of length {len(arr)} cannot be cut to length {length}')
    if len(arr) > length:
        raise ValueError(f'input to pad_vector_length() must be 1D, but got {arr.ndim}D array input')
    if len(arr) != length:
        arr = np.concatenate([arr, np.repeat(fill, length - len(arr))])
    elif copy:
        return arr.copy()
    return arr


def check_array(arr, dtype=None, ndim=None, shape=None, name=None):
    """
    Throw an exception if array does not conform to the provided criteria.

    Parameters
    ----------
    arr : array_like
        Array to check.
    dtype : np.dtype
        Required base dtype.
    ndim : int
        Required array dimensionality.
    shape : tuple of int
        Required array shape.
    name : str
        Name of checked array for optional debugging purposes.
    """
    if name is None:
        name = 'array'

    def list_string(lst):
        if len(lst) == 1:
            return str(lst[0])
        elif len(lst) == 2:
            return ' or '.join(map(str, lst))
        else:
            return ', '.join(map(str, lst[:-1])) + ', or ' + str(lst[-1])

    if dtype is not None:
        dtypes = [dtype] if np.isscalar(dtype) else dtype
        if not any(np.issubdtype(dt, arr.dtype) for dt in dtypes):
            reqs = list_string(dtypes)
            raise ValueError(f'{name} must have dtype {reqs}, but got {arr.dtype}')

    if ndim is not None:
        ndims = [ndim] if np.isscalar(ndim) else ndim
        if not any(nd == arr.ndim for nd in ndims):
            reqs = list_string(ndims)
            raise ValueError(f'{name} must be {reqs} dimensional, but got {arr.ndim} ndims')

    if shape is not None:
        if np.isscalar(shape):
            shape = [shape]
        shapes = [shape] if np.isscalar(shape[0]) else shape
        shapes = [tuple(s) for s in shapes]
        if not any(s == arr.shape for s in shapes):
            reqs = list_string(shapes)
            raise ValueError(f'{name} must have shape {reqs}, but got shape {arr.shape}')



import os
import numpy as np

from surfa.core.framed import FramedArray


class Overlay(FramedArray):

    def __init__(self, data, **kwargs):
        super().__init__(basedim=1, data=data, **kwargs)


def cast_overlay(obj, allow_none=True, copy=False):
    """
    Cast object to `Overlay` type.

    Parameters
    ----------
    obj : any
        Object to cast.
    allow_none : bool
        Allow for `None` to be successfully passed and returned by cast.
    copy : bool
        Return copy if object is already the correct type.

    Returns
    -------
    Overlay or None
        Casted overlay.
    """
    if obj is None and allow_none:
        return obj

    if isinstance(obj, Overlay):
        return obj.copy() if copy else obj

    if getattr(obj, '__array__', None) is not None:
        return Volume(np.array(obj))

    raise ValueError('cannot convert type %s to overlay' % type(obj).__name__)



import os
import sys
import platform
import subprocess as sp


def run(command, silent=False, background=False, executable='/bin/bash', log=None):
    """
    Runs a shell command and returns the exit code.

    Parameters
    ----------
    command : str
        Command to run.
    silent : bool
        Send output to devnull.
    background : bool
        Run command as a background process.
    executable : str
        Shell executable. Defaults to bash.
    log : str
        Send output to a log file.

    Returns
    -------
    int
        Command exit code.
    """

    # redirect the standard output appropriately
    if silent:
        std = {'stdout': sp.DEVNULL, 'stderr': sp.DEVNULL}
    elif not background:
        std = {'stdout': sp.PIPE, 'stderr': sp.STDOUT}
    else:
        std = {}  # do not redirect

    # run the command
    process = sp.Popen(command, **std, shell=True, executable=executable)
    if not background:
        # write the standard output stream
        if process.stdout:
            for line in process.stdout:
                decoded = line.decode('utf-8')
                if log is not None:
                    with open(log, 'a') as file:
                        file.write(decoded)
                sys.stdout.write(decoded)
        # wait for process to finish
        process.wait()

    return process.returncode


def collect_output(command, executable='/bin/bash'):
    """
    Collects the output of a shell command.

    Parameters
    ----------
    command : str
        Command to run.
    executable : str
        Shell executable. Defaults to bash.

    Returns
    -------
    tuple of (str, int)
        Tuple containing the command output and the corresponding exit code.
    """
    result = sp.run(command, stdout=sp.PIPE, stderr=sp.STDOUT, shell=True, executable=executable)
    return (result.stdout.decode('utf-8'), result.returncode)


def hostname(short=True):
    """
    Gets the system hostname.

    Parameters
    ----------
        short: Provide the short hostname. Defaults to True.
    """
    node = platform.node()
    if short:
        return node.split('.')[0]
    return node


def vmpeak():
    """
    Returns the peak memory usage of the process in kilobytes.

    Note: This only works on linux machines because itf requires `/proc/self/status`.
    """
    # TODO: switch to this (portable across platforms)
    # return resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
    procstatus = '/proc/self/status'
    if os.path.exists(procstatus):
        with open(procstatus, 'r') as file:
            for line in file:
                if 'VmPeak' in line:
                    return int(line.split()[1])
    return None


def fatal(message, retcode=1):
    """
    Prints an error message and exits or raises an exception if in interactive mode.

    Parameters
    ----------
    message : str
        Error message to print
    retcode : int
        Exit code. Defaults to 1.
    """
    import __main__ as main
    if hasattr(main, '__file__'):
        print(f'Error: {message}')
        sys.exit(retcode)
    else:
        raise Exception(message)



import os
import numpy as np

from surfa import Mesh
from surfa.transform import ImageGeometry
from surfa.io import fsio
from surfa.io import protocol
from surfa.io.utils import read_bytes
from surfa.io.utils import write_bytes
from surfa.io.utils import read_int
from surfa.io.utils import write_int
from surfa.io.utils import check_file_readability


def load_mesh(filename, fmt=None):
    """
    Load a mesh file into a Mesh object. Supported file formats are:

      - 'fs': Default FreeSurfer surface format.
      - 'gifti': Gifti geometry format.

    Parameters
    ----------
    filename : str
        File path to read.
    fmt : str, optional
        Forced file format. If None (default), file format is extrapolated
        from extension or assumed to be in the extensionless 'fs' surface format.

    Returns
    -------
    Mesh
        A mesh object loaded from file.
    """
    check_file_readability(filename)

    if fmt is None:
        iop = find_mesh_protocol_by_extension(filename)
    else:
        iop = protocol.find_protocol_by_name(mesh_io_protocols, fmt)

    return iop().load(filename)


def save_mesh(mesh, filename, fmt=None):
    """
    Save a Mesh object to file.

    Parameters
    ----------
    mesh : Mesh
        Mesh object to write.
    filename: str
        Destination file path.
    fmt : str
        Forced file format. If None (default), file format is extrapolated
        from extension or assumed to be in the extensionless 'fs' surface format.
    """
    if fmt is None:
        iop = find_mesh_protocol_by_extension(filename)
    else:
        iop = protocol.find_protocol_by_name(mesh_io_protocols, fmt)
        filename = iop.enforce_extension(filename)

    iop().save(mesh, filename)


def find_mesh_protocol_by_extension(filename):
    """
    Find mesh IO protocol from file extension. Since FS mesh files
    have no default extension (sadly), this function wraps and customizes
    `find_protocol_by_name` for mesh files.

    Parameters
    ----------
    filename : str
        File path to read.

    Returns
    -------
    protocol : IOProtocol
        Matched mesh IO protocol class.
    """

    # manually check for common, invalid file types that
    # people might naively try to load as meshes
    bad_formats = ('.mgh', '.mgz', '.nii', '.nii.gz')
    basename = os.path.basename(filename).lower()
    if basename.endswith(bad_formats):
        raise ValueError(f'{basename} is not a valid mesh file type.')

    # find matching protocol
    iop = protocol.find_protocol_by_extension(mesh_io_protocols, filename)

    # assume an unmatched extension is an FS surface
    if iop is None:
        iop = FreesurferSurfaceIO

    return iop


class FreesurferSurfaceIO(protocol.IOProtocol):
    """
    Mesh IO protocol for FreeSurfer-type files.
    """

    name = 'fs'
    extensions = ('.srf', '.surf')

    def load(self, filename):
        """
        Load a FreeSurfer surface file into a Mesh object.

        Parameters
        ----------
        filename : str
            File path to read.

        Returns
        -------
        Mesh
            A Mesh object loaded from file.
        """
        with open(filename, 'rb') as file:

            magic = read_int(file, size=3)
            if magic != -2:
                raise NotImplementedError(f'version {magic} FS surfaces cannot be loaded')

            # skip over two lines (second newline should follow immediately)
            file.readline(200)
            file.readline()

            # read mesh data
            nvertices = read_int(file)
            nfaces = read_int(file)

            vertices = np.fromfile(file, dtype='>f4', count=nvertices * 3).reshape((nvertices, 3))
            faces = np.fromfile(file, dtype='>i4', count=nfaces * 3).reshape((nfaces, 3))

            # create surface object
            mesh = Mesh(vertices, faces)

            # read metadata tags
            while True:
                tag, length = fsio.read_tag(file)
                if tag is None:
                    break

                # command history
                elif tag == fsio.tags.history:
                    history = file.read(length).decode('utf-8').rstrip('\x00')
                    if mesh.metadata.get('history'):
                        mesh.metadata['history'].append(history)
                    else:
                        mesh.metadata['history'] = [history]

                # read surface geometry
                elif tag == fsio.tags.old_surf_geom:
                    # read 8 hardcoded lines
                    string = ''.join([file.readline().decode('utf-8') for _ in range(8)])
                    mesh.geom = fsio.image_geometry_from_string(string)

                # real ras
                elif tag in (fsio.tags.old_real_ras, fsio.tags.real_ras):
                    if read_int(file) != 0:
                        mesh.metadata['real-ras'] = True

                # skip everything else
                else:
                    file.read(length)

        return mesh

    def save(self, mesh, filename):
        """
        Save a Mesh object to a FreeSurfer surface file.

        Parameters
        ----------
        mesh : Mesh
            Mesh to save.
        filename : str
            Destination file path.
        """
        mesh = mesh.convert(space='surface', copy=False)
        
        with open(filename, 'wb') as file:

            # magic and two lines for header
            write_int(file, -2, size=3)
            file.write('\n\n'.encode('utf-8'))

            # write topology
            write_int(file, mesh.nvertices)
            write_int(file, mesh.nfaces)

            # todo ensure contiguous
            mesh.vertices.astype('>f4').tofile(file)
            mesh.faces.astype('>i4').tofile(file)

            # geometry
            fsio.write_tag(file, fsio.tags.old_surf_geom)
            file.write(fsio.image_geometry_to_string(mesh.geom).encode('utf-8'))

            # real ras metadata
            fsio.write_tag(file, fsio.tags.old_real_ras)
            write_int(file, int(mesh.metadata.get('real-ras', False)))

            # history metadata
            for hist in mesh.metadata.get('history', []):
                fsio.write_tag(file, fsio.tags.history, len(hist))
                file.write(hist.encode('utf-8'))


class GiftiIO(protocol.IOProtocol):
    """
    Mesh IO protocol for gifti files.
    """

    name = 'gifti'
    extensions = ('.gii', '.gii.gz')

    def __init__(self):
        try:
            import nibabel as nib
        except ImportError:
            raise ImportError('the nibabel python package must be installed for GIFTI surface IO')
        self.nib = nib

    def load(self, filename):
        """
        Load a gifti mesh file into a Mesh object.

        Parameters
        ----------
        filename : str
            File path to read.

        Returns
        -------
        mesh : Mesh
            A Mesh object loaded from file.
        """
        gii = self.nib.load(filename)

        # extract mesh data
        vertices = gii.agg_data('pointset')
        faces = gii.agg_data('triangle')

        # ensure data exists
        if not vertices:
            raise RuntimeError('cannot load mesh since GIFTI file does not contain pointset data')
        if not faces:
            raise RuntimeError('cannot load mesh since GIFTI file does not contain triangle data')

        # extract geometry metadata
        metadata = gii.get_arrays_from_intent('pointset')[0].metadata
        shape = [
            metadata.get('VolGeomWidth'),
            metadata.get('VolGeomHeight'),
            metadata.get('VolGeomDepth'),
        ]
        voxsize = [
            metadata.get('VolGeomXsize'),
            metadata.get('VolGeomYsize'),
            metadata.get('VolGeomZsize'),
        ]
        rotation = [
            metadata.get('VolGeomX_R'),
            metadata.get('VolGeomY_R'),
            metadata.get('VolGeomZ_R'),
            metadata.get('VolGeomX_A'),
            metadata.get('VolGeomY_A'),
            metadata.get('VolGeomZ_A'),
            metadata.get('VolGeomX_S'),
            metadata.get('VolGeomY_S'),
            metadata.get('VolGeomZ_S'),
        ]
        center = [
            metadata.get('VolGeomC_R'),
            metadata.get('VolGeomC_A'),
            metadata.get('VolGeomC_S'),
        ]

        # assume that any missing geometry fields means invalid geometry
        if None in center + rotation + voxsize + shape:
            geometry = None
        else:
            shape = [int(i) for i in shape]
            voxsize = [float(i) for i in voxsize]
            rotation = np.reshape([float(i) for i in rotation], (3, 3))
            center = [float(i) for i in center]
            geometry = ImageGeometry(shape=shape, voxsize=voxsize, rotation=rotation, center=center)

        return Mesh(vertices, faces, geometry=geometry)

    def save(self, mesh, filename):
        """
        Save a Mesh object to a giti mesh file.

        Parameters
        ----------
        mesh : Mesh
            Mesh to save.
        filename : str
            Destination file path.
        """
        mesh = mesh.convert(space='surface', copy=False)

        gii = self.nib.GiftiImage()

        if mesh.geom is not None:
            meta = {
                'VolGeomWidth': '%d' % mesh.geom.shape[0],
                'VolGeomHeight': '%d' % mesh.geom.shape[1],
                'VolGeomDepth': '%d' % mesh.geom.shape[2],
                'VolGeomXsize': '%.15e' % mesh.geom.voxsize[0],
                'VolGeomYsize': '%.15e' % mesh.geom.voxsize[1],
                'VolGeomZsize': '%.15e' % mesh.geom.voxsize[2],
                'VolGeomX_R': '%.15e' % mesh.geom.rotation[0, 0],
                'VolGeomY_R': '%.15e' % mesh.geom.rotation[0, 1],
                'VolGeomZ_R': '%.15e' % mesh.geom.rotation[0, 2],
                'VolGeomX_A': '%.15e' % mesh.geom.rotation[1, 0],
                'VolGeomY_A': '%.15e' % mesh.geom.rotation[1, 1],
                'VolGeomZ_A': '%.15e' % mesh.geom.rotation[1, 2],
                'VolGeomX_S': '%.15e' % mesh.geom.rotation[2, 0],
                'VolGeomY_S': '%.15e' % mesh.geom.rotation[2, 1],
                'VolGeomZ_S': '%.15e' % mesh.geom.rotation[2, 2],
                'VolGeomC_R': '%.15e' % mesh.geom.center[0],
                'VolGeomC_A': '%.15e' % mesh.geom.center[1],
                'VolGeomC_S': '%.15e' % mesh.geom.center[2],
            }
        else:
            meta = None

        vertices = self.nib.gifti.GiftiDataArray(mesh.vertices, 'pointset', meta=meta)
        faces = self.nib.gifti.GiftiDataArray(mesh.faces, 'triangle')

        gii.add_gifti_data_array(vertices)
        gii.add_gifti_data_array(faces)
        self.nib.gifti.write(gii, filename)


# enabled mesh IO protocol classes
mesh_io_protocols = [
    FreesurferSurfaceIO,
    GiftiIO,
]



import copy
import warnings
import numpy as np

from surfa import transform
from surfa.core.array import check_array
from surfa.transform.space import cast_space


class Affine:

    def __init__(self, matrix, source=None, target=None, space=None):
        """
        N-D linear transform, represented by an (N, N) affine matrix, that
        tracks source and target geometries as well as coordinate space.

        Parameters
        ----------
        matrix : (N, N) or (N, N+1) float
            2D or 3D linear transformation matrix.
        source : castable ImageGeometry
            Transform source (or moving) image geometry.
        target : castable ImageGeometry
            Transform target (or fixed) image geometry.
        space : Space
            Coordinate space of the transform.
        """
        # init internal parameters
        self._matrix = None
        self._space = None
        self._source = None
        self._target = None

        # set the actual values
        self.matrix = matrix        
        self.space = space
        self.source = source
        self.target = target

    @property
    def _writeable(self):
        """
        Internal flag to determine whether or not the affine data can be
        modified. This is useful for cases like image geometries, which
        are represented by derivative affines that should not be changed.
        """
        return self.matrix.flags.writeable

    @_writeable.setter
    def _writeable(self, value):
        self.matrix.flags.writeable = value

    def _check_writeability(self):
        """
        Internal utility for ensuring that the affine is set as writeable.
        """
        if not self._writeable:
            raise ValueError('affine is read-only, use affine.copy() to duplicate before editing')

    @property
    def matrix(self):
        """
        The (N, N) affine matrix array.
        """
        return self._matrix
    
    @matrix.setter
    def matrix(self, mat):
        # check writeable
        if self.matrix is not None:
            self._check_writeability()

        # check input shape
        mat = np.ascontiguousarray(mat, dtype='float')
        check_array(mat, ndim=2, name='affine matrix')
        ndim = mat.shape[-1] - 1
        if ndim not in (2, 3) or mat.shape[0] not in (ndim, ndim + 1):
            raise ValueError('an N-D affine must be initialized with matrix of shape (N, N+1) '
                             f'or (N+1, N+1), where N is 2 or 3, but got shape {mat.shape}')

        # conform to square matrix
        square = np.eye(ndim + 1)
        square[: square.shape[0], :] = mat
        self._matrix = square

    def __array__(self):
        """
        Ensure affines can be cleanly casted to np.ndarray
        """
        return self.matrix

    @property
    def space(self):
        """
        Coordinate space of the transform.
        """
        return self._space
    
    @space.setter
    def space(self, value):
        self._check_writeability()
        self._space = cast_space(value, copy=True)

    @property
    def source(self):
        """
        Source (or moving) image geometry.
        """
        return self._source
    
    @source.setter
    def source(self, value):
        self._check_writeability()
        self._source = transform.cast_image_geometry(value, copy=True)

    @property
    def target(self):
        """
        Target (or fixed) image geometry.
        """
        return self._target
    
    @target.setter
    def target(self, value):
        self._check_writeability()
        self._target = transform.cast_image_geometry(value, copy=True)

    def save(self, filename):
        """
        Save affine transform to file.

        Parameters
        ----------
        filename : str
            Target filename.
        """
        from surfa.io.affine import save_affine
        save_affine(self, filename)

    def copy(self):
        """
        Return a copy of the affine.
        """
        aff = copy.deepcopy(self)
        aff._writeable = True
        return aff

    @property
    def ndim(self):
        """
        Dimensionality of the transform.
        """
        return self.matrix.shape[-1] - 1

    def __matmul__(self, other):
        """
        Matrix multiplication of affine matrices.
        """
        other = cast_affine(other, allow_none=False)
        if self.ndim != other.ndim:
            raise ValueError(f'cannot multiply {self.ndim}D and {other.ndim}D affines together')
        matrix = np.matmul(self.matrix, other.matrix)
        return Affine(matrix)

    def __call__(self, points):
        """
        Apply the affine transform matrix to a set of points. Calls `self.transform(points)`
        under the hood.
        """
        return self.transform(points)

    def transform(self, points):
        """
        Apply the affine transform matrix to an N-D point or set of points.

        Parameters
        ----------
        points : (..., N) float
            N-D point values to transform.

        Returns
        -------
        (..., N) float
            Transformed N-D point array.
        """
        points = np.asarray(points)
        if points.ndim == 1:
            points = points[np.newaxis]
        points = np.c_[points, np.ones(points.shape[0])].T
        return np.ascontiguousarray(np.dot(self.matrix, points).T.squeeze()[..., :-1])

    def inv(self):
        """
        Compute the inverse linear transform, reversing source and target information.

        Returns
        -------
        Affine
            Inverted affine transform.
        """
        inv = np.linalg.inv(self.matrix)
        return Affine(inv, source=self.target, target=self.source, space=self.space)

    def det(self):
        """
        Compute the affine matrix determinant.

        Returns
        -------
        float
            Determinant.
        """
        return np.linalg.det(self.matrix)

    def decompose(self, degrees=True):
        """
        Decompose the affine matrix into a set of translation, rotation, scale
        and shear components.

        Parameters
        ----------
        degrees : bool
            Return rotation in degrees instead of radians.

        Returns
        -------
        tuple
            Tuple of (translation, rotation, scale, shear) transform components.
        """
        translation = self.matrix[: self.ndim, -1]

        q, r = np.linalg.qr(self.matrix[: self.ndim, : self.ndim])

        di = np.diag_indices(self.ndim)
        scale = np.abs(r[di])

        p = np.eye(self.ndim)
        p[di] = r[di] / scale

        rotation = rotation_matrix_to_angles(q @ p, degrees=degrees)

        shear_mat = (p @ r) / np.expand_dims(scale, -1)
        if self.ndim == 2:
            shear = shear_mat[0, 1]
        else:
            shear = np.array([shear_mat[0, 1], shear_mat[0, 2], shear_mat[1, 2]])

        return (translation, rotation, scale, shear)

    def convert(self, source=None, target=None, space=None, copy=True):
        """
        Convert affine for a new coordinate space. Since original geometry and coordinate
        information is required for conversion, an error will be thrown if the source,
        target, and space have not been defined for the transform.

        Parameters
        ----------
        source, target : ImageGeometry, Image, or Surface
            Transform source and target image geometries.
        source : castable ImageGeometry
            Transform source (or moving) image geometry.
        target : castable ImageGeometry
            Transform target (or fixed) image geometry.
        space : Space
            Coordinate space of the transform.
        copy : bool
            Return copy of object if conditions are already satisfied.

        Returns
        -------
        float
            Converted affine transform.
        """
        source = self.source if source is None else transform.cast_image_geometry(source)
        target = self.target if target is None else transform.cast_image_geometry(target)
        space = self.space if space is None else cast_space(space)

        same_space = space == self.space
        same_source = transform.geometry.image_geometry_equal(source, self.source)
        same_target = transform.geometry.image_geometry_equal(target, self.target)

        # just return self if no changes are needed
        if all((same_space, same_source, same_target)):
            return self.copy() if copy else self

        # can't convert if we don't have original coordinate information
        if self.source is None or self.target is None or self.space is None:
            raise RuntimeError('original coordinate space, source, and target '
                               'information must be defined for affine conversion')

        if same_source and same_target:
            # just a simple conversion of transform coordinate space, without
            # changing source and target information
            a = self.target.affine(self.space, space)
            b = self.source.affine(space, self.space)
            affine = a @ self @ b
        else:
            # if source and target info is changing, we need to recompute the
            # transform by first converting it to universal world-space
            if self.space == 'world':
                affine = self.copy()
            else:
                a = self.target.affine(self.space, 'world')
                b = self.source.affine('world', self.space)
                affine = a @ self @ b
            # now convert into the desired coordinate space
            if space != 'world':
                a = target.affine('world', space)
                b = source.affine(space, 'world')
                affine = a @ affine @ b

        # the matrix multiplication above should propogate space and geometry info
        # correctly, but in a few cases, the matmul is skipped, so let's just return
        # a new affine object with the correct information
        return Affine(affine.matrix, source=source, target=target, space=space)


def affine_equal(a, b, matrix_only=False, tol=0.0):
    """
    Test whether two affine transforms are equivalent.

    Parameters
    ----------
    a, b : Affine
        Input affine transforms.
    matrix_only : bool
        Only compare matrix data, ignoring source and target information.
    tol : float
        Absolute error tolerance between affine matrices.

    Returns
    -------
    bool
        Returns True if the affines are equal.
    """
    try:
        a = cast_affine(a, allow_none=False)
        b = cast_affine(b, allow_none=False)
    except ValueError:
        return False

    if not np.allclose(a.matrix, b.matrix, atol=tol, rtol=0.0):
        return False

    if matrix_only:
        return True

    if not transform.geometry.image_geometry_equal(a.source, b.source) or \
       not transform.geometry.image_geometry_equal(a.target, b.target):
        return False

    if a.space != b.space:
        return False

    return True


def cast_affine(obj, allow_none=True, copy=False):
    """
    Cast object to `Affine` transform.

    Parameters
    ----------
    obj : any
        Object to cast.
    allow_none : bool
        Allow for `None` to be successfully passed and returned by cast.
    copy : bool
        Return copy of object if type already matches.

    Returns
    -------
    Affine or None
        Casted affine transform.
    """
    if obj is None and allow_none:
        return obj

    if isinstance(obj, Affine):
        return obj.copy() if copy else obj

    if getattr(obj, '__array__', None) is not None:
        return Affine(np.array(obj))

    raise ValueError('cannot convert type %s to affine' % type(obj).__name__)


def identity(ndim=3, **kwargs):
    """
    Identity affine transform.

    Parameters
    ----------
    ndim : int
        Dimensionality of affine transform. Can be 2 or 3.
    **kwargs
        Keyword arguments that are passed to `Affine` constructor.

    Returns
    -------
    Affine
        Identity affine transform.
    """
    return Affine(np.eye(ndim + 1), **kwargs)


def compose_affine(
    translation=None,
    rotation=None,
    scale=None,
    shear=None,
    ndim=3,
    degrees=True,
    **kwargs,
):
    """
    Compose an affine matrix from a set of N-D translation, rotation, scale, and shear
    transform components.

    Parameters
    ----------
    translation : float array
        N translation parameters.
    rotation : float array
        1 (2D) or 3 (3D) rotation angles.
    scale : float array
        N scale parameters.
    shear : float array
        1 (2D) or 3 (3D) shear parameters.
    ndim : int
        Dimensionality of transform.
    degrees : bool
        Define rotation in degrees instead of radians.
    **kwargs
        Keyword arguments that are passed to `Affine` constructor.

    Returns
    -------
    Affine
        Composed affine transform.
    """
    if ndim not in (2, 3):
        raise ValueError(f'affine transform must be 2D or 3D, got ndim {ndim}')

    if translation is None:
        translation = np.zeros(ndim)
    translation = np.asarray(translation)
    check_array(translation, shape=ndim, name='translation')

    if rotation is None:
        rotation = np.zeros(3) if ndim == 3 else np.zeros(1)
    if np.isscalar(rotation) and ndim == 2:
        rotation = [rotation]
    rotation = np.asarray(rotation)
    check_array(rotation, shape=(3 if ndim == 3 else 1), name='rotation')

    if scale is None:
        scale = np.ones(ndim)
    elif np.isscalar(scale):
        scale = np.repeat(scale, ndim).astype('float32')
    scale = np.asarray(scale)
    check_array(scale, shape=ndim, name='scale')

    if shear is None:
        shear = np.zeros(3) if ndim == 3 else np.zeros(1)
    if np.isscalar(shear) and ndim == 2:
        shear = [shear]
    shear = np.asarray(shear)
    check_array(shear, shape=(3 if ndim == 3 else 1), name='shear')

    T = np.eye(ndim + 1)
    T[:ndim, -1] = translation

    R = np.eye(ndim + 1)
    R[:ndim, :ndim] = angles_to_rotation_matrix(rotation, degrees=degrees)

    Z = np.diag(np.append(scale, 1))

    S = np.eye(ndim + 1)
    S[0][1] = shear[0]
    if ndim == 3:
        S[0][2] = shear[1]
        S[1][2] = shear[2]

    matrix = T @ R @ Z @ S

    return Affine(matrix, **kwargs)


def rotation_matrix_to_angles(matrix, degrees=True):
    """
    Compute rotation angle(s) from an (N, N) rotation matrix.

    Parameters
    ----------
    matrix : (N, N) float
        N-D rotation matrix.
    degrees : bool
        Return rotation angles in degrees instead of radians.

    Returns
    -------
    scalar or (3) float:
        Rotation angle(s).
    """
    ndim = matrix.shape[0]

    # matrix must be square
    check_array(matrix, ndim=2, name='rotation matrix')
    if matrix.shape[0] != matrix.shape[1]:
        raise ValueError('N-D rotation matrix must be square')

    if ndim == 2:
        rotation = np.arctan2(matrix[1][0], matrix[1][1])
    elif ndim == 3:
        rotation = np.array(
            [
                np.arctan2(matrix[1][2], matrix[2][2]),
                np.arctan2(
                    matrix[0][2], np.sqrt(matrix[1][2] ** 2 + matrix[2][2] ** 2.0)
                ),
                np.arctan2(matrix[0][1], matrix[0][0]),
            ]
        )
    else:
        raise ValueError('expected (N, N) rotation matrix, where N is 2 or 3, '
                        f'but got N of {ndim}')

    if degrees:
        rotation = np.degrees(rotation)

    return rotation


def angles_to_rotation_matrix(rotation, degrees=True):
    """
    Compute an (N, N) rotation matrix from a set of N-D rotation angles.

    Parameters
    ----------
    rotation : float scalar or array
        1 (2D) or 3 (3D) rotation angles.
    degrees : bool
        Define rotation in degrees instead of radians.

    Returns
    -------
    matrix : (N, N) float
        N-D rotation matrix.
    """
    if degrees:
        rotation = np.radians(rotation)

    # scalar value allowed for 2D transforms
    if np.isscalar(rotation):
        rotation = [rotation]
    num_angles = len(rotation)

    if num_angles == 1:
        c, s = np.cos(rotation[0]), np.sin(rotation[0])
        matrix = np.array([[c, -s], [s, c]], dtype='float')
    elif num_angles == 3:
        c, s = np.cos(rotation[0]), np.sin(rotation[0])
        rx = np.array([[1, 0, 0], [0, c, s], [0, -s, c]], dtype='float')
        c, s = np.cos(rotation[1]), np.sin(rotation[1])
        ry = np.array([[c, 0, s], [0, 1, 0], [-s, 0, c]], dtype='float')
        c, s = np.cos(rotation[2]), np.sin(rotation[2])
        rz = np.array([[c, s, 0], [-s, c, 0], [0, 0, 1]], dtype='float')
        matrix = rx @ ry @ rz
    else:
        raise ValueError(f'expected 1 (2D) or 3 (3D) rotation angles, got {num_angles}')

    return matrix


def random_affine(
    translation_range=0,
    rotation_range=0,
    scale_range=1,
    shear_range=0,
    ndim=3,
    degrees=True,
    **kwargs,
):
    """
    Draw a random affine from ranges of transform components. Affine will
    be identity by default.

    Parameters
    ----------
    translation_range : float scalar or array
        Range to sample translation parameters from. Scalar values define the max
        deviation from 0.0, while a 2-element array defines the (min, max) range.
    rotation_range : scalar or array
        Range to sample rotation parameters from. Scalar values define the max
        deviation from 0.0, while a 2-element array defines the (min, max) range.
    scale_range : scalar or array
        Range to sample scale parameters from. Scalar values define the max
        deviation from 1.0, while a 2-element array defines the (min, max) range.
    shear_range : scalar or array
        Range to sample shear parameters from. Scalar values define the max
        deviation from 0.0, while a 2-element array defines the (min, max) range.
    ndim : int
        Dimensionality of target transform.
    degrees : bool
        Define rotation in degrees instead of radians.
    **kwargs
        Keyword arguments that are passed to `Affine` constructor.

    Returns
    -------
    Affine
        Random affine transform.
    """

    # convert scalars to min/max ranges
    if np.isscalar(translation_range):
        translation_range = sorted([-translation_range, translation_range])
    if np.isscalar(rotation_range):
        rotation_range = sorted([-rotation_range, rotation_range])
    if np.isscalar(scale_range):
        scale_range = sorted([1.0 / scale_range, scale_range])
    if np.isscalar(shear_range):
        shear_range = sorted([-shear_range, shear_range])

    # conform to arrays
    translation_range = np.asarray(translation_range)
    rotation_range = np.asarray(rotation_range)
    scale_range = np.asarray(scale_range)
    shear_range = np.asarray(shear_range)

    # check ranges
    check_array(translation_range, shape=2, name='translation range')
    check_array(rotation_range, shape=2, name='rotation range')
    check_array(scale_range, shape=2, name='scale range')
    check_array(shear_range, shape=2, name='shear range')

    # compose from random paramters
    aff = compose_affine(
        translation=np.random.uniform(*translation_range, size=ndim),
        rotation=np.random.uniform(*rotation_range, size=(1 if ndim == 2 else 3)),
        scale=np.random.uniform(*scale_range, size=ndim),
        shear=np.random.uniform(*shear_range, size=(1 if ndim == 2 else 3)),
        ndim=ndim,
        degrees=degrees,
        **kwargs,
    )
    return aff



import numpy as np

from surfa.io import utils as iou
from surfa.core.labels import LabelLookup
from surfa.transform.geometry import ImageGeometry
from surfa.transform.geometry import cast_image_geometry


# FreeSurfer tag ID lookup
class tags:
    old_colortable = 1
    old_real_ras = 2
    history = 3
    real_ras = 4
    colortable = 5
    gcamorph_geom = 10
    gcamorph_type = 11
    gcamorph_labels = 12
    old_surf_geom = 20
    surf_geom = 21
    old_xform = 30
    xform = 31
    group_avg_area = 32
    auto_align = 33
    scalar_double = 40
    pedir = 41
    mri_frame = 42
    fieldstrength = 43


# these are FreeSurfer tags that have a
# buffer with hardcoded length
lengthless_tags = [
    tags.old_surf_geom,
    tags.old_real_ras,
    tags.old_colortable,
]


def read_tag(file):
    """
    Reads the next FreeSurfer tag ID and associated byte-length (if any)
    from a file buffer.

    Parameters
    ----------
    file : BufferedReader
        Opened file buffer to read tags from.

    Returns
    -------
    tag : int
        Tag ID read from file. Tag is None if end-of-file is reached.
    length : int
        Byte-length of tagged data.
    """
    tag = iou.read_int(file)
    if tag == 0:
        return (None, None)
    if tag == tags.old_xform:
        # backwards compatibility for transform, which writes 32-bit length
        length = iou.read_int(file)
    elif tag in lengthless_tags:
        # these tags have static lengths
        length = 0
    else:
        length = iou.read_int(file, size=8)
    return (tag, length)


def write_tag(file, tag, length=None):
    """
    Writes a FreeSurfer tag ID and associated byte-length (if any) to a file buffer.

    Parameters
    ----------
    file : BufferedWriter
        Opened file buffer to write tags to.
    tag : int
        Tag ID to write.
    length : int
        Byte-length of tagged data. Optional depending on tag type.
    """
    iou.write_int(file, tag)
    if tag == tags.old_xform:
        # backwards compatibility for transform, which writes 32-bit length
        iou.write_int(file, length)
    elif tag in lengthless_tags:
        # these tags have static lengths
        pass
    elif length is not None:
        iou.write_int(file, length, size=8)


def read_binary_lookup_table(file):
    """
    Read LabelLookup from a binary file buffer.

    Parameters
    ----------
    file : BufferedReader
        Opened file buffer to read labels from.

    Returns
    -------
    LabelLookup
        Labels read from file.
    """
    version = iou.read_bytes(file, '>i4')
    max_id = iou.read_bytes(file, '>i4')
    file_name_size = iou.read_bytes(file, '>i4')
    file.read(file_name_size).decode('utf-8')
    
    total = iou.read_bytes(file, '>i4')
    if total < 1:
        return None
    
    labels = LabelLookup()
    for n in range(total):
        index = iou.read_bytes(file, '>i4')
        name_size = iou.read_bytes(file, '>i4')
        name = file.read(name_size).decode('utf-8').rstrip('\x00')
        color = iou.read_bytes(file, '(4,)>i4').astype(np.float64)
        color[-1] = (255 - color[-1]) / 255
        labels[index] = (name, color)
    return labels


def write_binary_lookup_table(file, labels):
    """
    Write LabelLookup to a binary file buffer.

    Parameters
    ----------
    file : BufferedWriter
        Opened file buffer to write labels to.
    labels : LabelLookup
        Labels to write to file.
    """
    iou.write_bytes(file, -2, '>i4')
    iou.write_bytes(file, max(labels) + 1, '>i4')
    iou.write_bytes(file, 0, '>i4')
    iou.write_bytes(file, len(labels), '>i4')
    for index, element in labels.items():
        iou.write_bytes(file, index, '>i4')
        iou.write_bytes(file, len(element.name) + 1, '>i4')
        file.write((element.name + '\x00').encode('utf-8'))
        iou.write_bytes(file, element.color[:3].astype(np.uint8), '>i4')
        iou.write_bytes(file, (255 * (1 - element.color[-1])).astype(int), '>i4')


def image_geometry_from_string(string):
    """
    Convert ImageGeometry to a multi-line FS-style string.

    Parameters
    ----------
    string : str
       Image geometry in FS string format.

    Returns
    -------
    ImageGeometry
        Converted image geometry.
    """
    lines = string.splitlines()

    validline = lines[0].split()
    if validline[0] != 'valid':
        raise ValueError(f"geometry string must begin with 'valid' key, but got '{validline[0]}'")

    if validline[2] == '0':
        return None

    shape   = np.asarray(lines[2].split()[2:], dtype=np.int64)
    voxsize = np.asarray(lines[3].split()[2:], dtype=np.float64)
    xrot    = np.asarray(lines[4].split()[2:], dtype=np.float64)
    yrot    = np.asarray(lines[5].split()[2:], dtype=np.float64)
    zrot    = np.asarray(lines[6].split()[2:], dtype=np.float64)
    center  = np.asarray(lines[7].split()[2:], dtype=np.float64)

    rotation = np.stack([xrot, yrot, zrot], axis=1)

    return ImageGeometry(shape=shape, voxsize=voxsize, center=center, rotation=rotation)


def image_geometry_to_string(geom):
    """
    Convert ImageGeometry from a multi-line FS-style string.

    Parameters
    ----------
    geom : ImageGeometry
        Image geometry to convert.

    Returns
    -------
    str
        Image geometry in FS string format.
    """
    geom = cast_image_geometry(geom)

    valid = 1
    if geom is None:
        valid = 0
        geom = ImageGeometry(shape=(256, 256, 256))

    string  = 'valid = %d\n' % valid
    string += 'filename = none\n'
    string += 'volume = %d %d %d\n' % tuple(geom.shape)
    string += 'voxelsize = %.15e %.15e %.15e\n' % tuple(geom.voxsize)
    string += 'xras   = %.15e %.15e %.15e\n' % tuple(geom.rotation[:, 0])
    string += 'yras   = %.15e %.15e %.15e\n' % tuple(geom.rotation[:, 1])
    string += 'zras   = %.15e %.15e %.15e\n' % tuple(geom.rotation[:, 2])
    string += 'cras   = %.15e %.15e %.15e\n' % tuple(geom.center)
    return string



import pytest
import numpy as np
import surfa as sf


def basedim_to_framed_type(basedim):
    atype = {
        # 1: sf.Overlay,
        2: sf.Slice,
        3: sf.Volume,
    }.get(basedim)
    if atype is None:
        raise ValueError(f'Unsupported FramedArray dimensions: {basedim}')
    return atype


def test_data_management():
    """
    TODOC
    """

    na = np.random.rand(16, 16, 16)
    fa = sf.Volume(na)
    assert na is fa.data

    fac = sf.Volume(fa)
    assert fac.data is fa.data

    fac = sf.Volume(fa.copy())
    assert fac.data is not fa.data

    fac = fa.copy()
    assert fac.data is not fa.data


def test_data_shapes():
    """
    TODOC
    """

    for dim in (2, 3):  # TODO make overlay...

        BaseClass = basedim_to_framed_type(dim)

        # TODOC
        baseshape = tuple([32] * dim)
        fa = BaseClass(np.zeros(baseshape))
        assert fa.basedim == dim
        assert fa.nframes == 1
        assert fa.shape == baseshape
        assert fa.baseshape == baseshape
        assert fa.data.shape == baseshape
        assert fa.framed_data.shape == (*baseshape, 1)

        # TODOC
        nframes = 4
        shape = (*baseshape, nframes)
        fa = BaseClass(np.zeros(shape))
        assert fa.basedim == dim
        assert fa.nframes == nframes
        assert fa.shape == shape
        assert fa.baseshape == baseshape
        assert fa.data.shape == shape
        assert fa.framed_data.shape == shape

        # TODOC
        for x in np.arange(1, dim):
            inshape = tuple([8] * x)
            fa = BaseClass(np.zeros(inshape))
            outshape = np.ones(dim)
            outshape[:x] = inshape
            outshape = tuple(outshape)
            assert fa.basedim == dim
            assert fa.nframes == 1
            assert fa.shape == outshape
            assert fa.baseshape == outshape

        # TODOC
        with pytest.raises(ValueError):
            BaseClass(np.asarray(0))

        # TODOC
        for x in np.arange(3):
            with pytest.raises(ValueError):
                BaseClass(np.zeros(tuple([8] * (dim + 2 + x))))


def test_operators_and_assignment():
    """
    Super tedious, low-level test to make sure python math operators mimic
    the numpy standard. If these tests fail, then there are some big problems.
    """

    def array_equal(a, b, expected=sf.core.framed_array.FramedArray):
        assert isinstance(a, expected)
        assert isinstance(b, np.ndarray)
        assert a.shape == b.shape
        assert a.dtype == b.dtype
        assert np.array_equal(a, b)

    # test for all possible dimensions
    # for dim in (1, 2, 3):
    for dim in (2, 3):

        BaseClass = basedim_to_framed_type(dim)

        # generate a random numpy array, copy it, and initialize a FramedArray.
        # anything done to the FramedArray should match the numpy array exactly.
        shape = tuple([32] * dim)
        na = np.random.rand(*shape)
        fa = BaseClass(na.copy())

        # unary operators
        array_equal(+fa, +na)
        array_equal(-fa, -na)

        # binary operators
        array_equal(fa ** 3, na ** 3)
        array_equal(fa + 0.5, na + 0.5)
        array_equal(fa - 0.5, na - 0.5)
        array_equal(fa * 0.5, na * 0.5)
        array_equal(fa / 0.5, na / 0.5)

        # reverse binary operators
        array_equal(0.5 + fa, 0.5 + na)
        array_equal(0.5 - fa, 0.5 - na)
        array_equal(0.5 * fa, 0.5 * na)
        array_equal(0.5 / fa, 0.5 / na)

        # binary operators with numpy array
        nr = np.random.rand(*na.shape)
        array_equal(fa + nr, na + nr)
        array_equal(fa - nr, na - nr)
        array_equal(fa * nr, na * nr)
        array_equal(fa / nr, na / nr)

        # binary operators with framed array
        fr = BaseClass(nr.copy())
        array_equal(fa + fr, na + nr)
        array_equal(fa - fr, na - nr)
        array_equal(fa * fr, na * nr)
        array_equal(fa / fr, na / nr)

        # reverse binary operators with numpy array
        array_equal(nr + fa, nr + fa, expected=np.ndarray)
        array_equal(nr - fa, nr - fa, expected=np.ndarray)
        array_equal(nr * fa, nr * fa, expected=np.ndarray)
        array_equal(nr / fa, nr / fa, expected=np.ndarray)

        x = na.flatten()[10]  # grab some arbitrary element
        scaling = np.random.rand(4)
        for factor in (1, nr, fr):

            # assignment operators
            fab = fa.copy()
            nab = na.copy()
            fab += scaling[0] * factor
            nab += scaling[0] * factor
            fab -= scaling[1] * factor
            nab -= scaling[1] * factor
            fab *= scaling[2] * factor
            nab *= scaling[2] * factor
            fab /= scaling[3] * factor
            nab /= scaling[3] * factor
            array_equal(fab, nab)

            # comparisons
            array_equal(fa == x, na == x)
            array_equal(fa != x, na != x)
            array_equal(fa <= x, na <= x)
            array_equal(fa >= x, na >= x)

        # more comparisons
        fab = (fa < 0.5) & (fa > 0.2) | (fa > 0.8)
        nab = (na < 0.5) & (na > 0.2) | (na > 0.8)
        array_equal(fab, nab)


def test_data_functions():
    """
    TODOC
    """

    fa = sf.Volume(np.random.rand(16, 16, 16, 2))

    assert fa.max() == fa.data.max()
    assert fa.min() == fa.data.min()
    assert fa.mean() == fa.data.mean()
    assert fa.min(nonzero=True) == fa.data[fa.data != 0].min()
    assert fa.mean(nonzero=True) == fa.data[fa.data != 0].mean()

    # TODO update for newer numpy
    p = [0.2, 0.4, 0.6, 0.8]
    for method in ('linear', 'midpoint'):
        assert np.array_equal(fa.percentile(p, method=method),
                              np.percentile(fa.data, p, interpolation=method))
        assert np.array_equal(fa.percentile(p, method=method, nonzero=True),
                              np.percentile(fa.data[fa.data != 0], p, interpolation=method))



import os
import numpy as np
import gzip

from surfa import Volume
from surfa import Slice
from surfa import Overlay
from surfa import ImageGeometry
from surfa.core import FramedArray
from surfa.core import pad_vector_length
from surfa.image import FramedImage
from surfa.io import fsio
from surfa.io import protocol
from surfa.io.utils import read_int
from surfa.io.utils import write_int
from surfa.io.utils import read_bytes
from surfa.io.utils import write_bytes
from surfa.io.utils import check_file_readability


def load_volume(filename, fmt=None):
    """
    Load an image `Volume` from a 3D array file.

    Parameters
    ----------
    filename : str
        File path to read.
    fmt : str, optional
        Explicit file format. If None (default), the format is extrapolated
        from the file extension.

    Returns
    -------
    Volume
        Loaded volume.
    """
    return load_framed_array(filename=filename, atype=Volume, fmt=fmt)


def load_slice(filename, fmt=None):
    """
    Load an image `Slice` from a 2D array file.

    Parameters
    ----------
    filename : str
        File path to read.
    fmt : str, optional
        Explicit file format. If None (default), the format is extrapolated
        from the file extension.

    Returns
    -------
    Slice
        Loaded slice.
    """
    return load_framed_array(filename=filename, atype=Slice, fmt=fmt)


def load_overlay(filename, fmt=None):
    """
    Load a surface `Overlay` from a 1D array file.

    Parameters
    ----------
    filename : str
        File path to read.
    fmt : str, optional
        Explicit file format. If None (default), the format is extrapolated
        from the file extension.

    Returns
    -------
    Overlay
        Loaded overlay.
    """
    return load_framed_array(filename=filename, atype=Overlay, fmt=fmt)


def load_framed_array(filename, atype, fmt=None):
    """
    Generic loader for `FramedArray` objects.

    Parameters
    ----------
    filename : str
        File path to read.
    atype : class
        Particular FramedArray subclass to read into.
    fmt : str, optional
        Forced file format. If None (default), file format is extrapolated
        from extension.

    Returns
    -------
    FramedArray
        Loaded framed array. 
    """
    check_file_readability(filename)

    if fmt is None:
        iop = protocol.find_protocol_by_extension(array_io_protocols, filename)
        if iop is None:
            if atype is Overlay:
                # some freesurfer overlays do not have file extensions (another bizarre convention),
                # so let's fallback to the 'curv' format here
                iop = FreeSurferCurveIO
            else:
                raise ValueError(f'cannot determine file format from extension for {filename}')
    else:
        iop = protocol.find_protocol_by_name(array_io_protocols, fmt)
        if iop is None:
            raise ValueError(f'unknown file format {fmt}')

    return iop().load(filename, atype)


def save_framed_array(arr, filename, fmt=None):
    """
    Save a `FramedArray` object to file.

    Parameters
    ----------
    arr : FramedArray
        Object to write.
    filename: str
        Destination file path.
    fmt : str
        Forced file format. If None (default), file format is extrapolated
        from extension.
    """
    if fmt is None:
        iop = protocol.find_protocol_by_extension(array_io_protocols, filename)
        if iop is None:
            raise ValueError(f'cannot determine file format from extension for {filename}')
    else:
        iop = protocol.find_protocol_by_name(array_io_protocols, fmt)
        if iop is None:
            raise ValueError(f'unknown file format {fmt}')
        filename = iop.enforce_extension(filename)

    iop().save(arr, filename)


class MGHArrayIO(protocol.IOProtocol):
    """
    Array IO protocol for MGH and compressed MGZ files.
    """

    name = 'mgh'
    extensions = ('.mgz', 'mgh', '.mgh.gz')

    def dtype_from_id(self, id):
        """
        Convert a FreeSurfer datatype ID to a numpy datatype.

        Parameters
        ----------
        id : int
            FreeSurfer datatype ID.

        Returns
        -------
        np.dtype
            Converted numpy datatype.
        """
        mgh_types = {
            0:  '>u1',  # uchar
            1:  '>i4',  # int32
            2:  '>i8',  # int64
            3:  '>f4',  # float
            4:  '>i2',  # short
            6:  '>f4',  # tensor
            10: '>u2',  # ushort
        }
        dtype = mgh_types.get(id)
        if dtype is None:
            raise NotImplementedError(f'unsupported MGH data type ID: {id}')
        return np.dtype(dtype)

    def load(self, filename, atype):
        """
        Read array from an MGH/MGZ file.

        Parameters
        ----------
        filename : str
            File path to read.
        atype : class
            FramedArray subclass to load.

        Returns
        -------
        FramedArray
            Array object loaded from file.
        """

        # check if the file is gzipped
        fopen = gzip.open if filename.lower().endswith('gz') else open
        with fopen(filename, 'rb') as file:

            # skip version tag
            file.read(4)

            # read shape and type info
            shape = read_bytes(file, '>u4', 4)
            dtype_id = read_bytes(file, '>u4')
            dof = read_bytes(file, '>u4')

            # read geometry
            geom_params = {}
            unused_header_space = 254
            valid_geometry = bool(read_bytes(file, '>u2'))

            # ignore geometry if flagged as invalid
            if valid_geometry:
                geom_params = dict(
                    voxsize=read_bytes(file, '>f4', 3),
                    rotation=read_bytes(file, '>f4', 9).reshape((3, 3), order='F'),
                    center=read_bytes(file, '>f4', 3),
                )
                unused_header_space -= 60

            # skip empty header space
            file.read(unused_header_space)

            # read data buffer (MGH files store data in fortran order)
            dtype = self.dtype_from_id(dtype_id)
            data = read_bytes(file, dtype, int(np.prod(shape))).reshape(shape, order='F')

            # init array
            arr = atype(data.squeeze())

            # read scan parameters
            scan_params = {
                'tr': read_bytes(file, dtype='>f4'),
                'fa': read_bytes(file, dtype='>f4'),
                'te': read_bytes(file, dtype='>f4'),
                'ti': read_bytes(file, dtype='>f4'),
            }

            # ignore fov
            fov = read_bytes(file, dtype='>f4')
 
            # update image-specific information
            if isinstance(arr, FramedImage):
                arr.geom.update(**geom_params)
                arr.metadata.update(scan_params)

            # read metadata tags
            while True:
                tag, length = fsio.read_tag(file)
                if tag is None:
                    break

                # command history
                elif tag == fsio.tags.history:
                    history = file.read(length).decode('utf-8').rstrip('\x00')
                    if arr.metadata.get('history'):
                        arr.metadata['history'].append(history)
                    else:
                        arr.metadata['history'] = [history]

                # embedded lookup table
                elif tag == fsio.tags.old_colortable:
                    arr.labels = fsio.read_binary_lookup_table(file)

                # phase encode direction
                elif tag == fsio.tags.pedir:
                    pedir = file.read(length).decode('utf-8').rstrip('\x00')
                    if pedir != 'UNKNOWN':
                        arr.metadata['phase-encode-direction'] = pedir

                # field strength
                elif tag == fsio.tags.fieldstrength:
                    arr.metadata['field-strength'] = read_bytes(file, dtype='>f4')

                # skip everything else
                else:
                    file.read(length)

        return arr

    def save(self, arr, filename):
        """
        Write array to a MGH/MGZ file.

        Parameters
        ----------
        arr : FramedArray
            Array to save.
        filename : str
            Target file path.
        """

        # determine whether to write compressed data
        if filename.lower().endswith('gz'):
            fopen = lambda f: gzip.open(f, 'wb', compresslevel=6)
        else:
            fopen = lambda f: open(f, 'wb')

        with fopen(filename) as file:

            # before we map dtypes to MGZ-supported types, smartly convert int64 to int32
            if arr.dtype == np.int64:
                if arr.max() > np.iinfo(np.int32).max or arr.min() < np.iinfo(np.int32).min:
                    raise ValueError('MGH files only support int32 datatypes, but array cannot be ',
                                     'casted since its values exceed the int32 integer limits')
                arr = arr.astype(np.int32)

            # determine supported dtype to save as (order here is very important)
            type_map = {
                np.uint8: 0,
                np.bool8: 0,
                np.int32: 1,
                np.floating: 3,
                np.int16: 4,
                np.uint16: 10,
            }
            dtype_id = next((i for dt, i in type_map.items() if np.issubdtype(arr.dtype, dt)), None)
            if dtype_id is None:
                raise ValueError(f'writing dtype {arr.dtype.name} to MGH format is not supported')

            # sanity check on the array size
            ndim = arr.data.ndim
            if ndim < 1:
                raise ValueError(f'cannot save scalar value to MGH file format')
            if ndim > 4:
                raise ValueError(f'cannot save array with more than 4 dims to MGH format, but got {ndim}D array')

            # shape must always be a length-4 vector, so let's pad with ones
            shape = np.ones(4)
            shape[:ndim] = arr.data.shape

            # begin writing header
            write_bytes(file, 1, '>u4')  # version
            write_bytes(file, shape, '>u4')  # shape
            write_bytes(file, dtype_id, '>u4')  # MGH data type
            write_bytes(file, 1, '>u4')  # DOF

            # write geometry, if valid
            unused_header_space = 254
            is_image = isinstance(arr, FramedImage)
            write_bytes(file, is_image, '>u2')
            if is_image:
                write_bytes(file, arr.geom.voxsize, '>f4')
                write_bytes(file, np.ravel(arr.geom.rotation, order='F'), '>f4')
                write_bytes(file, arr.geom.center, '>f4')
                unused_header_space -= 60

            # fill empty header space
            file.write(bytearray(unused_header_space))

            # write array data
            write_bytes(file, np.ravel(arr.data, order='F'), self.dtype_from_id(dtype_id))

            # write scan parameters
            write_bytes(file, arr.metadata.get('tr', 0.0), '>f4')
            write_bytes(file, arr.metadata.get('fa', 0.0), '>f4')
            write_bytes(file, arr.metadata.get('te', 0.0), '>f4')
            write_bytes(file, arr.metadata.get('ti', 0.0), '>f4')

            # compute FOV
            volsize = pad_vector_length(arr.baseshape, 3, 1)
            fov = max(arr.geom.voxsize * volsize) if is_image else arr.shape[0]
            write_bytes(file, fov, '>f4')

            # write lookup table tag
            if arr.labels is not None:
                fsio.write_tag(file, fsio.tags.old_colortable)
                fsio.write_binary_lookup_table(file, arr.labels)

            # phase encode direction
            pedir = arr.metadata.get('phase-encode-direction', 'UNKNOWN')
            fsio.write_tag(file, fsio.tags.pedir, len(pedir))
            file.write(pedir.encode('utf-8'))

            # field strength
            fsio.write_tag(file, fsio.tags.fieldstrength, 4)
            write_bytes(file, arr.metadata.get('field-strength', 0.0), '>f4')

            # write history tags
            for hist in arr.metadata.get('history', []):
                fsio.write_tag(file, fsio.tags.history, len(hist))
                file.write(hist.encode('utf-8'))


class NiftiArrayIO(protocol.IOProtocol):
    """
    Array IO protocol for nifti files.
    """
    name = 'nifti'
    extensions = ('.nii.gz', '.nii')

    def __init__(self):
        try:
            import nibabel as nib
        except ImportError:
            raise ImportError('the `nibabel` python package must be installed for nifti IO')
        self.nib = nib

    def load(self, filename, atype):
        """
        Read array from a nifiti file.

        Parameters
        ----------
        filename : str
            File path read.
        atype : class
            FramedArray subclass to load.

        Returns
        -------
        FramedArray
            Array object loaded from file.
        """
        nii = self.nib.load(filename)
        data = nii.get_data()
        arr = atype(data)
        if isinstance(arr, FramedImage):
            matrix = nii.get_affine()
            voxsize = nii.header['pixdim'][1:4]
            arr.geom.update(vox2world=matrix, voxsize=voxsize)
        return arr

    def save(self, arr, filename):
        """
        Write array to a nifti file.

        Parameters
        ----------
        arr : FramedArray
            Array to save.
        filename : str
            Target file path.
        """
        isimage = isinstance(arr, FramedImage)
        matrix = arr.geom.vox2world.matrix if isimage else np.eye(4)
        nii = self.nib.Nifti1Image(arr.data, matrix)
        if is_image:
            nii.header['pixdim'][1:4] = arr.voxsize
        self.nib.save(nii, filename)


class FreeSurferAnnotationIO(protocol.IOProtocol):
    """
    Array IO protocol for 1D mesh annotation files.
    """
    name = 'annot'
    extensions = '.annot'

    def labels_to_mapping(self, labels):
        """
        The annotation file format saves each vertex label value as a
        bit-manipulated int32 value that represents an RGB. But, the label
        lookup table is embedded in the file, so it's kind of a pointless
        format that could be simply replaced by an MGH file with embedded
        labels, like any other volumetric segmentaton. This function builds
        a mapping to convert label RGBs to a lookup of bit-manipulated int32
        values. Using this mapping, we can convert between a classic integer
        segmentation and annotation-style values.
        """
        rgb = np.array([elt.color[:3].astype(np.int32) for elt in labels.values()])
        idx = np.array(list(labels.keys()))
        mapping = np.zeros(idx.max() + 1, dtype=np.int32)
        mapping[idx] = (rgb[:, 2] << 16) + (rgb[:, 1] << 8) + rgb[:, 0]
        return mapping

    def load(self, filename, atype):
        """
        Read overlay from an annot file.

        Parameters
        ----------
        filename : str
            File path read.
        atype : class
            FramedArray subclass to load. When reading annot files, this
            must be Overlay.

        Returns
        -------
        Overlay
            Array object loaded from file.
        """
        if atype is not Overlay:
            raise ValueError('annotation files can only be loaded as 1D overlays')

        with open(filename, 'rb') as file:

            nvertices = read_bytes(file, '>i4')
            data = np.zeros(nvertices, dtype=np.int32)

            value_map = read_bytes(file, '>i4', nvertices * 2)
            vnos = value_map[0::2]
            vals = value_map[1::2]
            data[vnos] = vals

            tag, length = fsio.read_tag(file)
            if tag is None or tag != fsio.tags.old_colortable:
                raise ValueError('annotation file does not have embedded label lookup data')
            labels = fsio.read_binary_lookup_table(file)

        # cache the zero value annotations (unknown labels)
        unknown_mask = data == 0

        # conver annotation values to corresponding label values
        mapping = self.labels_to_mapping(labels)
        ds = np.argsort(mapping)
        pos = np.searchsorted(mapping[ds], data)
        index = np.take(ds, pos, mode='clip')
        mask = mapping[index] != data
        data = np.ma.array(index, mask=mask)

        # all of the unknown labels should be converted to -1
        data[unknown_mask] = -1

        return Overlay(data, labels=labels)

    def save(self, arr, filename):
        """
        Write overlay to an annot file.

        Parameters
        ----------
        arr : Overlay
            Array to save.
        filename : str
            Target file path.
        """
        if not isinstance(arr, Overlay):
            raise ValueError(f'can only save 1D overlays as annotations, but got array type {typle(arr)}')

        if not np.issubdtype(arr.dtype, np.integer):
            raise ValueError(f'annotations must have integer dtype, but overlay has dtype {arr.dtype}')

        if arr.nframes > 1:
            raise ValueError(f'annotations must only have 1 frame, but overlay has {arr.nframes} frames')

        if arr.labels is None:
            raise ValueError('overlay must have label lookup if saving as annotation')

        # 
        unknown_mask = arr.data < 0

        # make sure all indices exist in the label lookup
        cleaned = arr.data[np.logical_not(unknown_mask)]
        found = np.in1d(cleaned, list(arr.labels.keys()))
        if not np.all(found):
            missing = list(np.unique(cleaned[found == False]))
            raise ValueError('cannot save overlay as annotation because it contains the following values '
                            f'that do not exist in its label lookup: {missing}')

        cleaned = arr.data.copy()
        cleaned[unknown_mask] = 0
        colors = self.labels_to_mapping(arr.labels)[arr.data]
        colors[unknown_mask] = 0

        with open(filename, 'bw') as file:
            # write the total number of vertices covered by the overlay
            nvertices = arr.shape[0]
            write_bytes(file, nvertices, '>i4')

            # write the data as sequences of (vertex number, color) for every 'vertex'
            # in the annotation, where color is 
            annot = np.zeros(nvertices * 2, dtype=np.int32)
            annot[0::2] = np.arange(nvertices, dtype=np.int32)
            annot[1::2] = colors
            write_bytes(file, annot, '>i4')

            # include the label lookup information
            fsio.write_tag(file, fsio.tags.old_colortable)
            fsio.write_binary_lookup_table(file, arr.labels)


class FreeSurferCurveIO(protocol.IOProtocol):
    """
    Array IO protocol for 1D FS curv files. This is another silly file format that
    could very well just be replaced by MGH files.
    """
    name = 'curv'
    extensions = ()

    def load(self, filename, atype):
        """
        Read overlay from a curv file.

        Parameters
        ----------
        filename : str
            File path read.
        atype : class
            FramedArray subclass to load. When reading curv files, this
            must be Overlay.

        Returns
        -------
        Overlay
            Array object loaded from file.
        """
        if atype is not Overlay:
            raise ValueError('curve files can only be loaded as 1D overlays')
        with open(filename, 'rb') as file:
            magic = read_int(file, size=3)
            nvertices = read_bytes(file, '>i4')
            read_bytes(file, '>i4')
            read_bytes(file, '>i4')
            data = read_bytes(file, '>f4', nvertices)
        return Overlay(data)

    def save(self, arr, filename):
        """
        Write overlay to a curv file.

        Parameters
        ----------
        arr : Overlay
            Array to save.
        filename : str
            Target file path.
        """
        if arr.nframes > 1:
            raise ValueError(f'curv files must only have 1 frame, but overlay has {arr.nframes} frames')

        with open(filename, 'bw') as file:
            write_int(file, -1, size=3)
            write_bytes(file, arr.shape[0], '>i4')
            write_bytes(file, 0, '>i4')
            write_bytes(file, 1, '>i4')
            write_bytes(file, arr.data, '>f4')


class ImageSliceIO(protocol.IOProtocol):
    """
    Generic array IO protocol for common image formats.
    """

    def __init__(self):
        try:
            from PIL import Image
        except ImportError:
            raise ImportError(f'the `pillow` python package must be installed for {self.name} IO')
        self.Image = Image

    def save(self, arr, filename):
        self.Image.fromarray(arr.data).save(filename)

    def load(self, filename):
        image = np.asarray(self.Image.open(filename))
        return Slice(image)


class JPEGArrayIO(ImageSliceIO):
    name = 'jpeg'
    extensions = ('.jpeg', '.jpg')


class PNGArrayIO(ImageSliceIO):
    name = 'png'
    extensions = '.png'


class TIFFArrayIO(ImageSliceIO):
    name = 'tiff'
    extensions = ('.tif', '.tiff')


# enabled array IO protocol classes
array_io_protocols = [
    MGHArrayIO,
    NiftiArrayIO,
    FreeSurferAnnotationIO,
    FreeSurferCurveIO,
    JPEGArrayIO,
    PNGArrayIO,
    TIFFArrayIO,
]



# _____
# SURFA
#

__version__ = '0.0.7'

from . import system

from .core import LabelLookup

from .transform import Affine
from .transform import Space
from .transform import ImageGeometry

from .image import Volume
from .image import Slice

from .mesh import Mesh
from .mesh import Overlay

from .io import load_volume
from .io import load_slice
from .io import load_overlay
from .io import load_affine
from .io import load_label_lookup
from .io import load_mesh

from . import vis



class IOProtocol:
    """
    Abstract (and private) protocol class to implement filetype-specific reading and writing.

    Subclasses must override the `load()` and `save()` methods, and set the `name` and `extensions`
    global class members.
    """

    name = ''
    extensions = []

    @classmethod
    def primary_extension(cls):
        """
        Return the primary (first) file extension of the protocol.
        """
        if not cls.extensions:
            return ''
        elif isinstance(cls.extensions, str):
            return cls.extensions
        return cls.extensions[0]

    @classmethod
    def enforce_extension(cls, filename):
        """
        Enforce a valid protocol extension on a filename. Returns the corrected filename.
        """
        if filename.lower().endswith(cls.extensions):
            return filename
        ext = cls.primary_extension()
        if not ext.startswith("."):
            ext = "." + ext
        return filename + cls.primary_extension()

    def load(self, filename):
        """
        File load function to be implemented for each subclass.
        """
        raise NotImplementedError(f'loading {self.name} files is not supported')

    def save(self, obj, filename):
        """
        File save function to be implemented for each subclass.
        """
        raise NotImplementedError(f'saving {self.name} files is not supported')


def find_protocol_by_name(protocols, fmt):
    """
    Find IO protocol by format name.

    Parameters
    ----------
    protocols : list
        List of IOProtocol classes to search.
    fmt : str
        File format name.

    Returns
    -------
    protocol : IOProtocol
        Matched IO protocol class.
    """
    fmt = fmt.lower()
    return next((p for p in protocols if fmt == p.name), None)


def find_protocol_by_extension(protocols, filename):
    """
    Find IO protocol by extension type.

    Parameters
    ----------
    protocols : list
        List of IOProtocol classes to search.
    filename : str
        Filename to grab extension of.

    Returns
    -------
    protocol : IOProtocol
        Matched IO protocol class.
    """
    return next((p for p in protocols if filename.lower().endswith(p.extensions)), None)


def get_all_extensions(protocols):
    """
    Returns all extensions in a list of protocols.

    Parameters
    ----------
    protocols : list
        List of IOProtocol classes to search.

    Returns
    -------
    extensions : list
        List of extensions.
    """
    extensions = []
    for protocol in protocols:
        if isinstance(protocol.extensions, str):
            extensions.append(protocol.extensions)
        else:
            extensions.extend(protocol.extensions)
    return extensions



import collections
import numpy as np
from copy import deepcopy


class LabelElement:

    def __init__(self, name=None, color=None):
        """
        Base LabelLookup element to define label name and color.

        Parameters
        ----------
        name : str
            Label name
        color : array_like
            Label color indicated by RGB or RGBA array.
        """
        self.name = name
        self.color = color

    @property
    def name(self):
        """
        Label name.
        """
        return self._value

    @name.setter
    def name(self, value):
        self._value = '' if value is None else str(value)

    @property
    def color(self):
        """
        Label color stored as an RGBA array of type (uchar, uchar, uchar, float).
        """
        return self._color

    @color.setter
    def color(self, value):
        if value is None:
            self._color = [0, 0, 0]
        if len(value) == 3:
            value = (*color, 1.0)
        elif len(value) != 4:
            raise ValueError('label color must be a 4-element RGBA array')
        color = np.array(value, dtype=np.float64)
        color[:3] = color[:3].clip(0, 255).astype(np.uint8).astype(np.float64)
        self._color = color


class LabelLookup(collections.OrderedDict):
    """
    Dictionary storing a label lookup mapping integer indices to labels names and colors.
    """

    def __setitem__(self, key, value):
        if not np.issubdtype(type(key), np.integer):
            raise ValueError(f'cannot convert object of type {key.__class__.__name__} to LabelLookup integer index')
        if isinstance(value, LabelElement):
            value = deepcopy(value)
        elif isinstance(value, str):
            value = LabelElement(name=value)
        elif isinstance(value, tuple) or isinstance(value, list) and len(value) == 2:
            value = LabelElement(name=value[0], color=value[1])
        else:
            raise ValueError(f'cannot convert object of type {value.__class__.__name__} to LabelLookup element')
        return super().__setitem__(int(key), value)

    def __repr__(self):
        col1 = len(str(max(self.keys()))) + 1
        col2 = max([len(elt.name) for elt in self.values()]) + 2
        lines = []
        for idx, elt in self.items():
            rgb = elt.color[:3].astype(np.uint8).astype(str)
            color_str = ',  '.join([str(c).rjust(3) for c in rgb]) + f',  {elt.color[-1]:.2f}'
            lines.append(str(idx).ljust(col1) + elt.name.ljust(col2) + color_str)
        return '\n'.join(lines)

    def save(self, filename, fmt=None):
        """
        Write label lookup to file.

        Parameters
        ----------
        filename : str
            Target filename to write lookup to.
        """
        from surfa.io.labels import save_label_lookup
        save_label_lookup(self, filename, fmt)

    def search(self, name, exact=False):
        """
        Search for 

        Parameters
        ----------
        name : str
            String or substring to search for.
        extact : bool
            If enabled, label must match search name exactly.

        Returns
        -------
        int or list of int
            Matching label indices. If `exact`, returns single index or None.
            Otherwise, a list of matches are returned.
        """
        if exact:
            return next((idx for idx, elt in self.items() if name == elt.name), None)
        else:
            allcaps = name.upper()
            return [idx for idx, elt in self.items() if allcaps in elt.name.upper()]

    def extract(self, labels):
        """
        Extract a new LabelLookup from a list of label indices.

        Parameters
        ----------
        labels : array_like of int
            List of label indices to extract.

        Returns
        -------
        LabelLookup
            Label lookup with extracted label indices.
        """
        lookup = LabelLookup()
        for label in labels:
            elt = self.get(label)
            if elt is None:
                raise ValueError(f'index {label} does not exist in the LabelLookup')
            lookup.add(label, elt.name, elt.color)
        return lookup

    def copy_colors(self, lookup):
        """
        Copies colors of matching label indices from a source LabelLookup.

        Parameters
        ----------
        lookup : LabelLookup
            Label lookup to copy colors from.
        """
        for label in self.keys():
            elt = lookup.get(label)
            if elt is not None and elt.color is not None:
                self[label].color = elt.color

    def copy_names(self, lookup):
        """
        Copies names of matching label indices from a source LabelLookup.

        Parameters
        ----------
        lookup : LabelLookup
            Label lookup to copy names from.
        """
        for label in self.keys():
            elt = lookup.get(label)
            if elt is not None:
                self[label].name = elt.name



class Space:

    def __init__(self, name):
        """
        Coordinate space representation. Supported spaces are:

            - voxel: Voxel (or image) coordinate space.
            - world: Universal world-space generally represented in RAS orientation.
            - surface: Surface or mesh coordinate space, dependent on base image geometry.

        Parameters
        ----------
        name : str
            Name of coordinate space, case-insensitive.
        """
        name = name.lower()

        # world space, defaulted to RAS space
        if name in ('w', 'world', 'ras'):
            name = 'world'
        # surface or mesh space
        elif name in ('s', 'surf', 'surface', 'm', 'mesh'):
            name = 'surface'
        # voxel or image space
        elif name in ('i', 'image', 'img', 'v', 'vox', 'voxel'):
            name = 'voxel'
        else:
            raise ValueError(f'unknown space: {name}')

        self._name = name

    def __eq__(self, other):
        """
        Test whether two spaces are the same.
        """
        try:
            other = cast_space(other, allow_none=False)
        except ValueError:
            return False
        return self.name == other.name

    def __repr__(self):
        return f"sf.Space('{self.name}')"

    def __str__(self):
        return self.name

    def copy(self):
        """
        Create a copy of the space.
        """
        return Space(self.name)

    @property
    def name(self):
        """
        Primary coordinate system name.
        """
        return self._name


def cast_space(obj, allow_none=True, copy=False):
    """
    Cast object to coordinate `Space`.

    Parameters
    ----------
    obj : any
        Object to cast.
    allow_none : bool
        Allow for `None` to be successfully passed and returned by cast.

    Returns
    -------
    Space or None
        Casted coordinate space.
    """
    if obj is None and allow_none:
        return obj

    if isinstance(obj, str):
        return Space(obj)

    if isinstance(obj, Space):
        return obj.copy() if copy else obj

    raise ValueError('cannot convert type %s to Space object' % type(obj).__name__)



import os
import numpy as np

from surfa.io import fsio
from surfa.io import protocol
from surfa.io import check_file_readability
from surfa.transform import Affine


def load_affine(filename, fmt=None):
    """
    Load an `Affine` from file.

    Parameters
    ----------
    filename : str
        File path to read.
    fmt : str, optional
        Forced file format. If None (default), file format is extrapolated
        from extension.

    Returns
    -------
    Affine
        Loaded affine. 
    """
    check_file_readability(filename)

    if fmt is None:
        iop = protocol.find_protocol_by_extension(affine_io_protocols, filename)
        if iop is None:
            raise ValueError(f'cannot determine file format from extension for {filename}')
    else:
        iop = protocol.find_protocol_by_name(affine_io_protocols, fmt)
        if iop is None:
            raise ValueError(f'unknown file format {fmt}')

    return iop().load(filename)


def save_affine(aff, filename, fmt=None):
    """
    Save a `Affine` object to file.

    Parameters
    ----------
    aff : Affine
        Object to write.
    filename: str
        Destination file path.
    fmt : str
        Forced file format. If None (default), file format is extrapolated
        from extension.
    """
    if fmt is None:
        iop = protocol.find_protocol_by_extension(affine_io_protocols, filename)
        if iop is None:
            raise ValueError(f'cannot determine file format from extension for {filename}')
    else:
        iop = protocol.find_protocol_by_name(affine_io_protocols, fmt)
        if iop is None:
            raise ValueError(f'unknown file format {fmt}')
        filename = iop.enforce_extension(filename)

    iop().save(aff, filename)


class LinearTransformArrayIO(protocol.IOProtocol):
    """
    Affine IO protocol for LTA files.
    """

    name = 'lta'
    extensions = ('.lta',)

    def load(self, filename):
        """
        Read affine from an LTA file.

        Parameters
        ----------
        filename : str
            File path to read.

        Returns
        -------
        Affine
            Affine object loaded from file.
        """
        with open(filename, 'r') as file:
            lines = [line.rstrip() for line in file]
            lines = [line for line in lines if line and not line.startswith('#')]

        # determine the coodinate space
        space_id = int(lines[0].split()[2])
        space = {0: 'vox',
                 1: 'world',
                 3: 'surf'}.get(space_id, None)
        if space is None:
            raise ValueError(f'unknown affine LTA type ID: {space_id}')

        # read in the actual matrix data
        matrix = np.asarray([line.split() for line in lines[5:9]], dtype=np.float64)

        # read in source and target geometry (if valid)
        source = fsio.image_geometry_from_string('\n'.join(lines[10:18]))
        target = fsio.image_geometry_from_string('\n'.join(lines[19:27]))
        if source is None and target is None:
            space = None

        return Affine(matrix, source=source, target=target, space=space)

    def save(self, aff, filename):
        """
        Write affine to an LTA file.

        Parameters
        ----------
        aff : Affine
            Array to save.
        filename : str
            Target file path.
        """
        with open(filename, 'w') as file:

            # determine LTA coordinate space
            if aff.space is None or aff.space == 'vox':
                file.write('type      = 0 # LINEAR_VOX_TO_VOX\n')
            elif aff.space == 'world':
                file.write('type      = 1 # LINEAR_RAS_TO_RAS\n')
            elif aff.space == 'surf':
                file.write('type      = 3 # LINEAR_SURF_TO_SURF\n')
            else:
                raise NotImplementedError(f'cannot write coodinate space {aff.space} to LTA - this is a '
                                           'bug, not a user error')

            # this is all useless legacy information
            file.write('nxforms   = 1\n')
            file.write('mean      = 0.0000 0.0000 0.0000\n')
            file.write('sigma     = 1.0000\n')
            file.write('1 4 4\n')

            # write the actual matrix data
            file.write('%.15e %.15e %.15e %.15e\n' % tuple(aff.matrix[0]))
            file.write('%.15e %.15e %.15e %.15e\n' % tuple(aff.matrix[1]))
            file.write('%.15e %.15e %.15e %.15e\n' % tuple(aff.matrix[2]))
            file.write('%.15e %.15e %.15e %.15e\n' % tuple(aff.matrix[3]))

            # write source geometry (if any)
            file.write('src volume info\n')
            file.write(fsio.image_geometry_to_string(aff.source))

            # write target geometry (if any)
            file.write('dst volume info\n')
            file.write(fsio.image_geometry_to_string(aff.target))


# enabled affine IO protocol classes
affine_io_protocols = [
    LinearTransformArrayIO,
]
